import logging

from pfx.pfxcore.decorator import rest_view
from pfx.pfxcore.models import PFXUser
from pfx.pfxcore.views import PermsRestView

logger = logging.getLogger(__name__)


@rest_view("/users")
class UserRestView(PermsRestView):
    model = PFXUser
    fields = ['first_name', 'last_name', 'username']



from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured

from . import default_settings


class PFXSettings:
    def __getattr__(self, name):
        try:
            val = getattr(django_settings, name)
        except AttributeError:
            if hasattr(default_settings, name):
                val = getattr(default_settings, name)
            else:
                raise
        if name == "PFX_SECRET_KEY" and not val:
            raise ImproperlyConfigured(
                "The PFX_SECRET_KEY setting must not be empty.")
        return val


settings = PFXSettings()



from unittest.mock import MagicMock, patch

from django.core.exceptions import ValidationError
from django.test import TestCase
from django.test.utils import override_settings

from pfx.pfxcore.exceptions import APIError
from pfx.pfxcore.test import APIClient, TestAssertMixin
from tests.models import User


class BasicAPIErrorTest(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        pass

    @override_settings(DEBUG=False)
    def test_resource_not_found(self):
        response = self.client.get('/api/error/404')
        self.assertRC(response, 404)

    @override_settings(DEBUG=False)
    @patch('pfx.pfxcore.decorator.rest.logger', MagicMock())
    def test_error500(self):
        response = self.client.get('/api/error/500')
        self.assertRC(response, 500)

    def test_malformed_json(self):
        response = self.client.post(
            '/api/authors', '''{
                "first_name": "Arthur Charles",
                "last_name": "Clarke",
                "name_length": 1,
                "gender": "male,
                "slug": "arthur-c-clarke"
                }''')
        self.assertRC(response, 422)
        self.assertJE(
            response, 'message',
            "JSON Malformed Invalid control character "
            "at: line 5 column 33 (char 155)")

        User.objects.create_user(
            username='jrr.tolkien',
            email="jrr.tolkien@oxford.com",
            password='RIGHT PASSWORD',
            first_name='John Ronald Reuel',
            last_name='Tolkien',
        )

        self.client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD')
        response = self.client.post(
            '/api/auth/change-password', '''{
                "old_password": "Wrong stuffs",
                "new_password": "Wrong stuffs",
                }''')
        self.assertRC(response, 422)
        self.assertIn(self.get_val(response, 'message'), [
            "JSON Malformed Illegal trailing comma before end "
            "of object: line 3 column 47 (char 96)",  # python >= 3.13
            "JSON Malformed Expecting property name enclosed in "
            "double quotes: line 4 column 17 (char 114)"])  # python <= 3.12

    def test_api_error_from_validation_error(self):
        error = APIError("Default", validation_error=ValidationError(
            ValidationError("Message")))
        self.assertEqual(error.data['message'], "Message")

        # If no general message, default message is used.
        error = APIError("Default", validation_error=ValidationError(
            ValidationError(dict(field1="Message"))))
        self.assertEqual(error.data['message'], "Default")



from django.test import TestCase

from pfx.pfxcore.test import APIClient, TestAssertMixin
from tests.views import BaseRestView, IllegalPriorityRestView


class ViewDecoratorTest(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient()

    def test_method_not_allowed(self):
        response = self.client.put('/api/authors')
        self.assertRC(response, 405)

    def test_method_not_allowed_if_non_http_method(self):
        response = self.client.generic('ILLEGAL', '/api/authors')
        self.assertEqual(response.status_code, 405)

    def test_method_not_allowed_if_not_exists(self):
        response = self.client.get('/api/does-not-exists')
        self.assertRC(response, 405)

    def test_explicit_priority(self):
        # With default priority, string > untyped.
        response = self.client.get(
            '/api/authors/priority/dynamic')
        self.assertRC(response, 200)
        self.assertJE(response, 'value', 'str:dynamic')

        # With default priority, static string > dynamic.
        response = self.client.get(
            '/api/authors/priority/default')
        self.assertRC(response, 200)
        self.assertJE(response, 'value', 'static:default')

        # With default priority, string > static string.
        response = self.client.get(
            '/api/authors/priority/hello')
        self.assertRC(response, 200)
        self.assertJE(response, 'value', 'str:hello')

        # With default priority, path is used
        response = self.client.get(
            '/api/authors/priority/hello/there')
        self.assertRC(response, 200)
        self.assertJE(response, 'value', 'path:hello/there')

        # With default priority, static path > dynamic path
        response = self.client.get(
            '/api/authors/priority/default/path')
        self.assertRC(response, 200)
        self.assertJE(response, 'value', 'static:default_path')

        # With low priority, string is used.
        response = self.client.get(
            '/api/authors/priority/priority-less')
        self.assertRC(response, 200)
        self.assertJE(response, 'value', 'str:priority-less')

        # With high priority, priorized is used.
        response = self.client.get(
            '/api/authors/priority/priority-more')
        self.assertRC(response, 200)
        self.assertJE(response, 'value', 'static:more')

    def test_illegal_priority(self):
        with self.assertRaises(Exception):
            IllegalPriorityRestView.as_urlpatterns()

    def test_path_order(self):
        paths = [
            '/api/xxx/<int:param>',
            '/api/xxx/me',
            '/api/xxx/<param>/other/test',
            '/api/xxx/<param>',
            '/api/xxx/<int:param>/other',
            '/api/xxx/<param>/other/<param2>']
        sorted_paths = sorted(paths, key=lambda p: BaseRestView._path_order(
            p, dict(priority=0)), reverse=True)
        self.assertEqual(sorted_paths, [
            '/api/xxx/me',
            '/api/xxx/<int:param>/other',
            '/api/xxx/<int:param>',
            '/api/xxx/<param>/other/test',
            '/api/xxx/<param>/other/<param2>',
            '/api/xxx/<param>'])

    def test_path_order_path(self):
        paths = [
            '/api/xxx/<path:path>',
            '/api/xxx/<param>/test']
        sorted_paths = sorted(paths, key=lambda p: BaseRestView._path_order(
            p, dict(priority=0)), reverse=True)
        self.assertEqual(sorted_paths, [
            '/api/xxx/<param>/test',
            '/api/xxx/<path:path>'])



from .authentication_views import (
    AuthenticationView,
    ForgottenPasswordView,
    OtpEmailView,
    SendMessageTokenMixin,
    SignupView,
)
from .fields import VF, FieldType, ViewField, ViewModelField
from .filters_views import Filter, FilterGroup, ModelFilter
from .locale_views import LocaleRestView
from .media_rest_view_mixin import MediaPermsRestViewMixin, MediaRestViewMixin
from .ordered_rest_view_mixin import OrderedRestViewMixin
from .rest_views import (
    BaseRestView,
    BodyMixin,
    CreatePermsRestViewMixin,
    CreateRestViewMixin,
    DeletePermsRestViewMixin,
    DeleteRestViewMixin,
    DetailPermsRestViewMixin,
    DetailRestViewMixin,
    ListPermsRestViewMixin,
    ListRestViewMixin,
    ModelBodyMixin,
    ModelMixin,
    ModelResponseMixin,
    PermsRestView,
    RestView,
    SecuredRestViewMixin,
    SlugDetailRestViewMixin,
    SlugPermsDetailRestViewMixin,
    UpdatePermsRestViewMixin,
    UpdateRestViewMixin,
    resource_not_found,
)



from decimal import Decimal

from django.db import connection, models
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import include, path

from pfx.pfxcore import register_views
from pfx.pfxcore.decorator import rest_property, rest_view
from pfx.pfxcore.fields import DecimalField
from pfx.pfxcore.models import JSONReprMixin
from pfx.pfxcore.test import APIClient, TestAssertMixin
from pfx.pfxcore.views import RestView
from tests.views import FakeViewMixin


class TestDecimalModel(JSONReprMixin, models.Model):
    decimal = DecimalField(
        max_digits=10, decimal_places=5, json_decimal_places=2)
    percent = DecimalField(
        max_digits=10, decimal_places=4, percent=True)
    monetary = DecimalField(
        max_digits=10, decimal_places=5, json_decimal_places=2,
        currency="CHF")

    @rest_property(field=DecimalField(
        max_digits=10, decimal_places=5, json_decimal_places=2,
        currency="CHF"))
    def price_double(self):
        return Decimal(self.monetary * 2)

    class Meta:
        verbose_name = "TestModel"
        verbose_name_plural = "TestModels"
        ordering = ['pk']


@rest_view("/test-decimal-model")
class DecimalModelRestView(FakeViewMixin, RestView):
    default_public = True
    model = TestDecimalModel
    fields = ['decimal', 'percent', 'monetary', 'price_double']


urlpatterns = [
    path('api/', include(register_views(DecimalModelRestView))),
    path('api/', include('pfx.pfxcore.urls'))
]


@override_settings(ROOT_URLCONF=__name__)
class TestFieldsDecimal(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(TestDecimalModel)

    def test_decimal_meta(self):
        response = self.client.get('/api/test-decimal-model/meta')
        self.assertRC(response, 200)
        self.assertJE(response, 'fields.decimal.percent', False)
        self.assertJE(response, 'fields.decimal.currency', None)
        self.assertJE(response, 'fields.percent.percent', True)
        self.assertJE(response, 'fields.percent.currency', None)
        self.assertJE(response, 'fields.monetary.percent', False)
        self.assertJE(response, 'fields.monetary.currency', "CHF")
        self.assertJE(response, 'fields.price_double.percent', False)
        self.assertJE(response, 'fields.price_double.currency', "CHF")

    def test_decimal(self):
        t = TestDecimalModel.objects.create(
            decimal=3.14, percent=0.0810, monetary=100)
        t.save()
        t.refresh_from_db()
        self.assertEqual(t.decimal, Decimal('3.14000'))
        self.assertEqual(t.percent, Decimal('0.0810'))

        response = self.client.get(f'/api/test-decimal-model/{t.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'decimal', "3.14")
        self.assertJE(response, 'percent', "8.10")
        self.assertJE(response, 'monetary', "100.00")
        self.assertJE(response, 'price_double', "200.00")

        response = self.client.put(f'/api/test-decimal-model/{t.pk}', dict(
            decimal="3.12121",
            percent="7.7",
        ))
        self.assertRC(response, 200)
        self.assertJE(response, 'decimal', "3.12")
        self.assertJE(response, 'percent', "7.70")

        t.refresh_from_db()
        self.assertEqual(t.decimal, Decimal('3.12121'))
        self.assertEqual(t.percent, Decimal('0.0770'))



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class DateFormat(QueryParameter):
    """Date format option.

    * if `false` (default value), date fields will be returned
    as ISO 8601 string.
    * if `true`, date fields will be returned as a JSON object with
    2 attributes:
        * `value`: the ISO 8601 string
        * `formatted`: a string representation
        (short format using the request locale)
    """
    name = 'date_format'
    schema = dict(type='boolean', default=False)


register_global_parameter(DateFormat)



import logging

from django.core.exceptions import FieldDoesNotExist
from django.http import FileResponse, HttpResponse
from django.shortcuts import redirect
from django.utils.translation import gettext_lazy as _

from pfx.pfxcore.decorator import rest_api
from pfx.pfxcore.exceptions import APIError, NotFoundError
from pfx.pfxcore.fields import MediaField
from pfx.pfxcore.http import JsonResponse
from pfx.pfxcore.shortcuts import get_bool, settings
from pfx.pfxcore.storage.s3_storage import StorageException

from . import parameters
from .rest_views import ModelMixin

logger = logging.getLogger(__name__)


class MediaRestViewMixin(ModelMixin):
    """Extension mixin to manage media fields."""

    def _get_model_field(self, field):
        try:
            model_field = self.model._meta.get_field(field)
            if not isinstance(model_field, MediaField):
                raise NotFoundError  # pragma: no cover
        except FieldDoesNotExist:  # pragma: no cover
            raise NotFoundError
        return model_field

    @rest_api(
        "/<int:pk>/<str:field>/upload-url/<str:filename>", method="get",
        priority_doc=-8)
    def field_media_upload_url(self, pk, field, filename, *args, **kwargs):
        """Entrypoint for
        :code:`GET /<int:pk>/<str:field>/upload-url/<str:filename>` route.

        Get the upload URL for a media file.

        :param pk: The object pk
        :param field: The field name
        :param filename: The file name
        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Get upload URL
            description: |
                Get upload URL for a `MediaField` field.

                You can upload a file ont the received URL. When the upload
                query is done, you have to confirm the process with an
                update request (`PUT`) on {model}. The body of this request
                must contain the name of the `MediaField` with the contents
                of the `file` value in the response of this request.

                1. `GET /<int:pk>/<str:field>/upload-url/<str:filename>`
                    → `data`
                2. `PUT data.url aFile Content-Type: aFile.type`
                3. `PUT /<int:pk> {{field: data.file}}`
            parameters extras:
                pk: The {model} pk.
                field: The {model} field name. Must be the name of
                    a `MediaField` field.
                filename: The desired filename.
            responses:
                200:
                    description: The upload URL
                    content:
                        application/json:
                            schema:
                                properties:
                                    url:
                                        type: string
                                        format: uri
                                    file:
                                        type: object
                                        properties:
                                            name:
                                                type: string
                                            key:
                                                type: string
        """
        obj = self.get_object(pk=pk)
        mediaField = self._get_model_field(field)
        if mediaField.storage.direct:
            raise APIError("Unavailable for direct storage")
        try:
            res = self._get_model_field(field).get_upload_url(
                self.request, obj, filename)
        except StorageException as e:  # pragma: no cover
            logger.exception(e)
            raise APIError(_("Unexpected storage error", status=500))
        return JsonResponse(res)

    @rest_api(
        "/<int:pk>/<str:field>", method="get",
        parameters=[parameters.MediaRedirect], priority_doc=-9)
    def field_media_get(self, pk, field, *args, **kwargs):
        """Entrypoint for :code:`GET /<int:pk>/<str:field>` route.

        Get the download URL for a media file.

        :param pk: The object pk
        :param field: The field name
        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Get {model} file
            description: Get the URL for a media field file.
            parameters extras:
                pk: the {model} pk
                field: the {model} field name
            responses:
                200:
                    description: |
                        The file stream if storage is direct,
                        otherwise the file URL, only if `redirect` is `false`.
                    content:
                        application/json:
                            schema:
                                properties:
                                    url:
                                        type: string
                                        format: uri
                        application/octet-stream:
                            schema:
                                type: file
                302:
                    description: |
                        The redirect, for undirect storage
                        if `redirect` is `true`.
        """
        obj = self.get_object(pk=pk)
        mediaField = self._get_model_field(field)
        try:
            url = mediaField.get_url(self.request, obj)
        except StorageException as e:  # pragma: no cover
            logger.exception(e)
            raise APIError(_("Unexpected storage error", status=500))

        if mediaField.storage.direct:
            filename = getattr(obj, field).get('name')
            if settings.STORAGE_LOCAL_X_ACCEL_REDIRECT:
                response = HttpResponse()
                response["Content-Disposition"] = (
                    f"attachment; filename={filename}")
                response['X-Accel-Redirect'] = (
                    f"/filestore/{getattr(obj, field).get('key')}")
                return response
            response = FileResponse(
                open(mediaField.get_url(self.request, obj), 'rb'),
                as_attachment=True, filename=getattr(obj, field).get('name'))
            return response

        if get_bool(self.request.GET, 'redirect'):
            return redirect(url)
        return JsonResponse(dict(url=url))


class MediaPermsRestViewMixin(MediaRestViewMixin):
    """Extension mixin to check permissions."""

    def field_media_upload_url_perm(self, *args, **kwargs):
        return self.request.user.has_perm(*self.get_model_perms('change'))

    def field_media_get_perm(self, *args, **kwargs):
        return self.request.user.has_perm(*self.get_model_perms('view'))



import inspect
import json
import re
from pathlib import Path

from django.core.management.base import BaseCommand

from apispec import APISpec
from apispec.utils import deepupdate
from apispec.yaml_utils import load_operations_from_docstring

from pfx.pfxcore import __PFX_VIEWS__
from pfx.pfxcore.apidoc import __PARAMETERS__, __TAGS__, ParameterGroup
from pfx.pfxcore.settings import settings
from pfx.pfxcore.views import FilterGroup, ModelMixin

DEFAULT_TEMPLATE = dict(
    title="PFX API",
    version="1.0.0",
    openapi_version="3.0.2")
RE_PATH_PARAMS = re.compile(r'<(?:(\w+):)?(\w+)>')
DJANGO_OPENAPI_TYPES = {
    'str': 'string',
    'int': 'integer',
    'slug': 'string',
    'uuid': 'string',
    'path': 'string',
}


def from_django_path(path):
    return RE_PATH_PARAMS.sub(r'{\g<2>}', path)


def path_parameters(spec, path):
    def as_dict(data):
        return data if isinstance(data, dict) else dict(description=data)

    existings = {
        p.get('name', '#N/A') for p in spec.get('parameters', [])
        if isinstance(p, dict)}
    extras = spec.pop('parameters extras', {}).copy()
    for ptype, name in RE_PATH_PARAMS.findall(path):
        if name in existings:
            # Ignore path parameters that are manually described.
            continue
        yield deepupdate({
            'in': 'path',
            'name': name,
            'schema': {'type': DJANGO_OPENAPI_TYPES.get(ptype, 'string')},
            'required': True}, as_dict(extras.get(name, {})))


def global_parameters(spec, method):
    def extend_groups(params):
        for p in params:
            if issubclass(p, ParameterGroup):
                yield from p.parameters
            else:
                yield p

    existings = {
        p.get('name', '#N/A') for p in spec.get('parameters', [])
        if isinstance(p, dict)}
    for qp in extend_groups(method.rest_api_params):
        if qp.name == 'search':
            search = (
                spec['search'] if 'search' in spec else method.rest_api_search)
            if not search:
                continue
            if isinstance(search, str):
                yield qp.as_parameter(doc=search)
                continue
        if qp.name in existings:
            # Ignore path parameters that are manually described.
            continue
        yield qp.id()


def filters_parameters(view):
    def get_filters(src):
        if not hasattr(src, 'filters'):
            return
        for e in src.filters:
            if isinstance(e, FilterGroup):
                yield from get_filters(e)
            else:
                yield e

    from pfx.pfxcore.views.fields import FieldType
    for f in get_filters(view):
        if f.type == FieldType.ModelObject and f.related_model:
            ftype = FieldType.from_model_field(
                f.related_model._meta.pk.__class__)
        else:
            ftype = f.type
        op = (
            not f.filter_func_list and
            f" ({f.filter_func_and and 'and' or 'or'})" or "")
        yield {
            'in': 'query',
            'name': str(f.name),
            'schema': dict(
                type='array', default=f.defaults,
                items=dict(type=FieldType.to_apidoc(ftype))),
            'description': f"Filter by {f.label.lower()}{op}"}


def get_operations(view, url, auths, groups):
    order = dict(get=1, post=2, put=3, delete=4)
    for op, method_name in sorted(
            url['methods'].items(), key=lambda e: order.get(e[0], 99)):
        method = getattr(view, method_name)
        rest_doc = view.rest_doc.get((url['path'], op), {}).copy()
        rest_groups = (
            (method.rest_api_groups | set(rest_doc.pop('groups', []))) or
            {'default', })
        if groups and not (groups & rest_groups):
            continue
        doc = inspect.getdoc(method)
        vars = {}
        if issubclass(view, ModelMixin) and view.model:
            vars.update(
                model=view.model._meta.verbose_name.lower(),
                models=view.model._meta.verbose_name_plural.lower())
        if doc:
            doc = doc.format(**vars)
        spec = deepupdate(
            load_operations_from_docstring(doc).get(op, {}), rest_doc)
        if auths and not view()._is_public(
                method.rest_api_public, method_name):
            # If not public, add all defined security schemes
            for a in auths:
                spec.setdefault('security', []).append({a: []})
        spec.setdefault('summary', from_django_path(url['fixed_path']))
        tags = view.get_apidoc_tags()
        tags_d = {}
        for tag in tags:
            tags_d[tag.name] = tag
        spec.setdefault('tags', list(tags_d.keys()))
        __TAGS__.update(tags_d)
        parameters = spec.setdefault('parameters', [])
        parameters.extend(global_parameters(spec, method))
        parameters.extend(path_parameters(spec, url['path']))
        if method.rest_api_filters:
            parameters.extend(filters_parameters(view))
        for p in parameters:
            if 'name' in p and p['name'] == 'id':
                p['name'] = 'pk'
        if method.rest_api_request_schema:
            schema = getattr(view, method.rest_api_request_schema)
            spec_schema = spec.setdefault('requestBody', {})
            if spec_schema:
                spec_schema.update(deepupdate(
                    schema.to_ref(schema=True),
                    spec_schema.copy()))
            else:
                spec_schema.update(schema.to_ref())
        if method.rest_api_response_schema:
            schema = getattr(view, method.rest_api_response_schema)
            spec_schema = spec.setdefault('responses', {}).setdefault(200, {})
            if spec_schema:
                spec_schema.update(deepupdate(
                    schema.to_ref(schema=True), spec_schema.copy()))
            else:
                spec_schema.update(schema.to_ref())
        yield op, spec


def get_spec(groups):
    re_pk = re.compile(r'<(\w+:)?id>')
    spec = APISpec(**{**DEFAULT_TEMPLATE, **settings.PFX_OPENAPI_TEMPLATE})
    auths = spec.to_dict().get(
        'components', {}).get('securitySchemes', {}).keys()
    for parameter in __PARAMETERS__:
        spec.components.parameter(
            parameter.id(), parameter.location, parameter.as_parameter())
    for view in __PFX_VIEWS__:
        view.generate_schemas()
        for schema in view.get_apidoc_schemas():
            if schema.id() not in spec.components.schemas:
                spec.components.schema(schema.id(), schema.to_schema())
        for url in view.get_urls():
            url['fixed_path'] = re_pk.sub(r'<\g<1>pk>', url['path'])
            spec.path(
                path=from_django_path(url['fixed_path']),
                operations=dict(get_operations(view, url, auths, groups)))
    for key in sorted(__TAGS__.keys()):
        spec.tag(__TAGS__[key].to_dict())

    # If application use x-tagGroups, remove groups without exported tags:
    tag_groups = spec.options.get('x-tagGroups')
    if tag_groups:
        spec.options['x-tagGroups'] = list(filter(
            lambda tg: set(tg.get('tags', [])) & set(__TAGS__.keys()),
            tag_groups))
    return spec


class Command(BaseCommand):
    help = 'Generate OpenAPI documentation'

    def add_arguments(self, parser):
        parser.add_argument("groups", nargs="*", type=str)
        parser.add_argument(
            "-f", "--format", dest="format", choices=['json', 'yaml'],
            default="json", type=str)

    def handle(self, *args, groups=None, **options):
        spec = get_spec(set(groups))

        path = Path(
            settings.PFX_OPENAPI_PATH,
            f'{settings.PFX_OPENAPI_FILENAME}.{options["format"]}')
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w") as outfile:
            if options["format"] == 'yaml':
                outfile.write(spec.to_yaml())
            else:
                json.dump(spec.to_dict(), outfile, indent=2)

        self.stdout.write(self.style.SUCCESS(
            f"OpenAPI documentation generated: {path}"))



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class SubsetLimit(QueryParameter):
    """The maximum number of objects in `offset` mode.

    If the value of `subset` is not `"offset"`, this attribute will
    be ignored.

    If `limit` > `PFX_MAX_LIST_RESULT_SIZE`, `PFX_MAX_LIST_RESULT_SIZE`
    will be applied.
    """
    name = 'limit'
    schema = dict(type='integer', default=10)


register_global_parameter(SubsetLimit)



from io import StringIO

from django.core.management import call_command
from django.test import TestCase, override_settings

from apispec.yaml_utils import load_operations_from_docstring as from_doc

from pfx.pfxcore.management.commands.makeapidoc import (
    get_spec,
    path_parameters,
)
from pfx.pfxcore.test import TestAssertMixin
from tests.views import AuthorRestView


class ApiDocTest(TestAssertMixin, TestCase):
    def test_default_generation(self):
        spec = get_spec(set()).to_dict()
        self.assertEqual(spec['openapi'], "3.0.2")
        info = spec['info']
        self.assertEqual(info['title'], "PFX API")
        self.assertEqual(info['version'], "1.0.0")

    @override_settings(PFX_OPENAPI_TEMPLATE=dict(
        title="MyAPI",
        info=dict(description="A test API")))
    def test_default_customized_generation(self):
        spec = get_spec(set()).to_dict()
        self.assertEqual(spec['openapi'], "3.0.2")
        info = spec['info']
        self.assertEqual(info['title'], "MyAPI")
        self.assertEqual(info['version'], "1.0.0")
        self.assertEqual(info['description'], "A test API")

    @override_settings(PFX_OPENAPI_TEMPLATE=dict(
        title="MyAPI",
        info=dict(
            description="A test API"),
        components=dict(securitySchemes=dict(
            BearerAuth=dict(type='http', scheme='Bearer')))))
    def test_paths_generation(self):
        def assertMethods(paths, p, methods):
            self.assertEqual(set(paths[p].keys()), methods)

        spec = get_spec(set()).to_dict()
        paths = spec['paths']
        assertMethods(paths, '/authors', {'get', 'post'})
        assertMethods(paths, '/authors/{pk}', {'get', 'put', 'delete'})
        assertMethods(paths, '/authors/slug/{slug}', {'get'})
        assertMethods(paths, '/authors/cache/{pk}', {'get'})

        tags = spec['tags']
        tags_keys = [t['name'] for t in tags]
        self.assertIn("Authentication", tags_keys)
        self.assertIn("Author", tags_keys)
        self.assertIn("Book", tags_keys)
        self.assertIn("Book Type", tags_keys)
        self.assertIn("Locale", tags_keys)
        self.assertJE(paths['/authors'], 'get.tags', ["Author"])
        self.assertJE(paths['/authors'], 'post.tags', ["Author"])
        self.assertJE(paths['/auth/login'], 'post.tags', ["Authentication"])

        # Check security data
        self.assertJENotExists(paths, '/authors.get.security')
        self.assertJEExists(paths, '/private/authors.get.security')
        security = self.get_val(paths, '/private/authors.get.security')
        self.assertIn(dict(BearerAuth=[]), security)

        # Check a inherited get with default description
        get = self.get_val(paths, '/authors/{pk}.get')
        self.assertJE(get, 'summary', "Get author")
        # See comment in pfx/pfxcore/views/parameters/groups.py
        # self.assertJE(get, 'parameters.@0', {
        #     '$ref': "#/components/parameters/DateFormat"})
        self.assertJE(get, 'parameters.@0.in', "path")
        self.assertJE(get, 'parameters.@0.name', "pk")
        self.assertJE(get, 'parameters.@0.schema.type', "integer")
        self.assertJE(get, 'parameters.@0.required', True)
        self.assertJE(get, 'parameters.@0.description', "the author pk")
        # Check a inherited get with custom description
        self.assertJE(
            paths, '/authors-annotate/{pk}.get.summary',
            "Get custom author")
        # Check a slug
        get = self.get_val(paths, '/authors/slug/{slug}.get')
        self.assertJE(get, 'summary', "Get author by slug")
        # See comment in pfx/pfxcore/views/parameters/groups.py
        # self.assertJE(get, 'parameters.@0', {
        #     '$ref': "#/components/parameters/DateFormat"})
        self.assertJE(get, 'parameters.@0.in', "path")
        self.assertJE(get, 'parameters.@0.name', "slug")
        self.assertJE(get, 'parameters.@0.schema.type', "string")
        self.assertJE(get, 'parameters.@0.required', True)
        self.assertJE(get, 'parameters.@0.description', "the author slug name")

        # Check list parameters with auto generated filters
        get = self.get_val(paths, '/authors.get')
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/ListCount'})
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/ListItems'})
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/ListSearch'})
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/ListOrder'})
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/Subset'})
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/SubsetPage'})
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/SubsetPageSize'})
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/SubsetPageSubset'})
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/SubsetOffset'})
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/ListCount'})
        self.assertJIn(get, 'parameters', {
            '$ref': '#/components/parameters/SubsetLimit'})
        self.assertJIn(get, 'parameters',     {
            "in": "query",
            "name": "science_fiction",
            "schema": {
                "type": "array",
                "default": [],
                "items": {
                    "type": "boolean"
                }
            },
            "description": "Filter by science fiction (or)"
        })
        self.assertJIn(get, 'parameters', {
            "in": "query",
            "name": "heroic_fantasy",
            "schema": {
                "type": "array",
                "default": [],
                "items": {
                    "type": "boolean"
                }
            },
            "description": "Filter by heroic fantasy (or)"
        })
        self.assertJIn(get, 'parameters', {
            "in": "query",
            "name": "types",
            "schema": {
                "type": "array",
                "default": [],
                "items": {
                    "type": "number"
                }
            },
            "description": "Filter by types (or)"
        })
        self.assertJIn(get, 'parameters', {
            "in": "query",
            "name": "last_name",
            "schema": {
                "type": "array",
                "default": [],
                "items": {
                    "type": "boolean"
                }
            },
            "description": "Filter by last name (or)"
        })
        self.assertJIn(get, 'parameters', {
            "in": "query",
            "name": "first_name",
            "schema": {
                "type": "array",
                "default": [],
                "items": {
                    "type": "string"
                }
            },
            "description": "Filter by first name (or)"
        })
        self.assertJIn(get, 'parameters', {
            "in": "query",
            "name": "gender",
            "schema": {
                "type": "array",
                "default": [],
                "items": {
                    "type": "string"
                }
            },
            "description": "Filter by gender (or)"
        })
        self.assertJIn(get, 'parameters', {
            "in": "query",
            "name": "last_name_choices",
            "schema": {
                "type": "array",
                "default": [],
                "items": {
                    "type": "string"
                }
            },
            "description": "Filter by tolkien or asimov (or)"
        })

        schema_key = self.get_val(
            paths,
            '/authors/{pk}.get.responses.200.content.application/json.schema'
        )['$ref'].split('/')[-1]
        self.assertIn("Author", schema_key)
        props = spec['components']['schemas'][schema_key]['properties']
        self.assertJE(props, 'first_name.type', 'string')
        # Check an @rest_property field (readonly)
        self.assertJE(props, 'name_length.type', 'number')
        self.assertJE(props, 'name_length.readonly', True)
        self.assertJE(props, 'gender.type', 'string')
        self.assertJE(props, 'gender.enum', ['male', 'female'])
        self.assertJE(props, 'types.type', 'array')
        self.assertJE(props, 'types.items.type', 'object')
        self.assertJE(props, 'created_at.type', 'string')
        self.assertJE(props, 'created_at.format', 'date')

        schema_key = self.get_val(
            paths,
            '/authors.get.responses.200.content.application/json.schema'
        )['$ref'].split('/')[-1]
        self.assertIn("Author", schema_key)
        props = spec['components']['schemas'][schema_key]['properties']
        self.assertJE(props, 'items.type', 'array')
        self.assertJE(props, 'items.items.format', 'Author')
        self.assertSize(props, 'items.items.properties', 6)
        props = self.get_val(props, 'items.items.properties')
        self.assertJE(props, 'pk.type', 'number')
        self.assertJE(props, 'resource_name.type', 'string')
        self.assertJE(props, 'resource_slug.type', 'string')
        self.assertJE(props, 'first_name.type', 'string')
        self.assertJE(props, 'first_name.readonly', False)
        self.assertJE(props, 'last_name.type', 'string')
        self.assertJE(props, 'last_name.readonly', False)
        self.assertJE(props, 'gender.type', 'string')
        self.assertJE(props, 'gender.enum', ['male', 'female'])

        schema_key = self.get_val(
            paths,
            '/authors-annotate/{pk}.get.responses.200.content'
            '.application/json.schema'
        )['$ref'].split('/')[-1]
        # Check ModelObjectList fields
        props = spec['components']['schemas'][schema_key]['properties']
        self.assertJE(props, 'books.type', 'array')
        self.assertJE(props, 'books.items.type', 'object')
        self.assertJE(props, 'books.items.format', 'Book')
        props = self.get_val(props, 'books.items.properties')
        self.assertJE(props, 'pk.type', 'number')
        self.assertJE(props, 'resource_name.type', 'string')
        self.assertJE(props, 'resource_reference.type', 'string')
        self.assertJE(props, 'author_name.type', 'string')
        self.assertJE(props, 'author_gender.type', 'string')

        schema_key = self.get_val(
            paths,
            '/books-custom-author/{pk}.get.responses.200.content'
            '.application/json.schema'
        )['$ref'].split('/')[-1]
        self.assertIn("Book", schema_key)
        props = spec['components']['schemas'][schema_key]['properties']
        # Check VF field with alias
        self.assertJE(props, 'book_name.type', 'string')
        self.assertJE(props, 'book_name.readonly', False)
        # Check VF field with django lookup (author__last_name)
        self.assertJE(props, 'author_last_name.type', 'string')
        self.assertJE(props, 'author_last_name.readonly', True)
        # Check that json_repr docstring from VF is used
        self.assertJE(props, 'author.properties.hello.type', 'string')
        self.assertJE(props, 'author.properties.last_name.type', 'string')

        schema_key = self.get_val(
            paths,
            '/books/{pk}.get.responses.200.content.application/json.schema'
        )['$ref'].split('/')[-1]
        self.assertIn("Book", schema_key)
        props = spec['components']['schemas'][schema_key]['properties']
        # Check that json_repr docstring from model is used
        self.assertJE(props, 'author.properties.new_field.type', 'string')
        # Check MinutesDurationField fields json schema
        self.assertJE(props, 'read_time.type', 'object')
        self.assertJE(props, 'read_time.properties.minutes.type', 'number')
        self.assertJE(
            props, 'read_time.properties.clock_format.type', 'string')
        self.assertJE(
            props, 'read_time.properties.human_format.type', 'string')

        schema_key = self.get_val(
            paths,
            '/books/meta.get.responses.200.content.application/json.schema'
        )['$ref'].split('/')[-1]
        self.assertIn("form_meta", schema_key)
        props = spec['components']['schemas'][schema_key]['properties']
        # Check that schema contains fields
        self.assertJE(props, 'fields.type', 'array')

        schema_key = self.get_val(
            paths,
            '/books/meta/list.get.responses.200.content'
            '.application/json.schema'
        )['$ref'].split('/')[-1]
        self.assertIn("list_meta", schema_key)
        props = spec['components']['schemas'][schema_key]['properties']
        # Check that schema contains fields+filters+orders
        self.assertJE(props, 'fields.type', 'array')
        self.assertJE(props, 'filters.type', 'array')
        self.assertJE(props, 'orders.type', 'array')

        schema_key = self.get_val(
            paths,
            '/books.post.requestBody.content.application/json.schema'
        )['$ref'].split('/')[-1]
        self.assertIn("Book", schema_key)
        props = spec['components']['schemas'][schema_key]['properties']
        # Check readonly is not in schema
        self.assertJENotExists(props, 'created_at')
        # Check foreign object can be number/object
        self.assertJE(props, 'author.oneOf.@0.type', 'number')
        self.assertJE(props, 'author.oneOf.@1.type', 'object')
        self.assertSize(props, 'author.oneOf.@1.properties', 1)
        self.assertJE(props, 'author.oneOf.@1.properties.pk.type', 'number')

    def test_paths_order(self):
        spec = get_spec(set()).to_dict()
        paths = self.get_val(spec, 'paths')
        authors_paths = [p for p in paths if p.startswith('/authors')]
        self.assertEqual(authors_paths[:11], [
            '/authors',
            '/authors/{pk}',
            '/authors/slug/{slug}',
            '/authors/cache/{pk}',
            '/authors/priority/{value}',
            '/authors/priority/default',
            '/authors/priority/default/path',
            '/authors/priority/priority-less',
            '/authors/priority/priority-more',
            '/authors/meta/list',
            '/authors/meta'])
        books_paths = [p for p in paths if p.startswith('/books')]
        self.assertEqual(books_paths, [
            "/books",
            "/books/{pk}",
            "/books/{pk}/{field}",
            "/books/{pk}/{field}/upload-url/{filename}",
            "/books/meta/list",
            "/books/meta",
            "/books-custom-author",
            "/books-custom-author/{pk}",
            "/books-custom-author/{pk}/{field}",
            "/books-custom-author/{pk}/{field}/upload-url/{filename}",
            "/books-custom-author/meta/list",
            "/books-custom-author/meta"])

    def test_groups(self):
        spec = get_spec(set()).to_dict()
        self.assertJEExists(spec, 'paths./authors/cache/{pk}.get')
        self.assertJEExists(spec, 'paths./authors-annotate/{pk}.get')

        spec = get_spec({'cache', }).to_dict()
        self.assertJEExists(spec, 'paths./authors/cache/{pk}.get')
        self.assertJENotExists(spec, 'paths./authors-annotate/{pk}.get')

        spec = get_spec({'custom', }).to_dict()
        self.assertJENotExists(spec, 'paths./authors/cache/{pk}.get')
        self.assertJEExists(spec, 'paths./authors-annotate/{pk}.get')

        spec = get_spec({'cache', 'custom'}).to_dict()
        self.assertJEExists(spec, 'paths./authors/cache/{pk}.get')
        self.assertJEExists(spec, 'paths./authors-annotate/{pk}.get')

        spec = get_spec({'default', }).to_dict()
        self.assertJENotExists(spec, 'paths./authors/cache/{pk}.get')
        self.assertJENotExists(spec, 'paths./authors-annotate/{pk}.get')

    def test_view_get_urls(self):
        def assertMethods(urls, p, methods):
            self.assertEqual(next(filter(
                lambda u: u['path'] == p, urls))['methods'], methods)

        urls = AuthorRestView.get_urls()

        # Methods from RestView
        assertMethods(urls, '/authors', dict(get='get_list', post='post'))
        assertMethods(urls, '/authors/<int:id>', dict(
            delete='delete', get='get', put='put'))
        # A method from SlugDetailRestViewMixin
        assertMethods(urls, '/authors/slug/<slug:slug>', dict(
            get='get_by_slug'))
        # A method from AuthorRestView itself
        assertMethods(urls, '/authors/cache/<int:id>', dict(get='cache_get'))

    def test_path_parameter_untyped(self):
        param = next(path_parameters(from_doc(""), '/path/<my_param>'))
        self.assertJE(param, 'in', "path")
        self.assertJE(param, 'name', "my_param")
        self.assertJE(param, 'schema.type', "string")
        self.assertJE(param, 'required', True)
        self.assertJENotExists(param, 'description')

    def test_path_parameter_str(self):
        param = next(path_parameters(from_doc(""), '/path/<str:my_param>'))
        self.assertJE(param, 'in', "path")
        self.assertJE(param, 'name', "my_param")
        self.assertJE(param, 'schema.type', "string")
        self.assertJE(param, 'required', True)
        self.assertJENotExists(param, 'description')

    def test_path_parameter_int(self):
        param = next(path_parameters(from_doc(""), '/path/<int:my_param>'))
        self.assertJE(param, 'in', "path")
        self.assertJE(param, 'name', "my_param")
        self.assertJE(param, 'schema.type', "integer")
        self.assertJE(param, 'required', True)
        self.assertJENotExists(param, 'description')

    def test_path_parameter_slug(self):
        param = next(path_parameters(from_doc(""), '/path/<slug:my_param>'))
        self.assertJE(param, 'in', "path")
        self.assertJE(param, 'name', "my_param")
        self.assertJE(param, 'schema.type', "string")
        self.assertJE(param, 'required', True)
        self.assertJENotExists(param, 'description')

    def test_path_parameter_uuid(self):
        param = next(path_parameters(from_doc(""), '/path/<uuid:my_param>'))
        self.assertJE(param, 'in', "path")
        self.assertJE(param, 'name', "my_param")
        self.assertJE(param, 'schema.type', "string")
        self.assertJE(param, 'required', True)
        self.assertJENotExists(param, 'description')

    def test_path_parameter_path(self):
        param = next(path_parameters(from_doc(""), '/path/<path:my_param>'))
        self.assertJE(param, 'in', "path")
        self.assertJE(param, 'name', "my_param")
        self.assertJE(param, 'schema.type', "string")
        self.assertJE(param, 'required', True)
        self.assertJENotExists(param, 'description')

    def test_path_parameter_custom(self):
        """Test with a custom path type.
        Actually you cannot register your custom path type. Every custom
        type will be considered as string."""
        param = next(path_parameters(from_doc(""), '/path/<custom:my_param>'))
        self.assertJE(param, 'in', "path")
        self.assertJE(param, 'name', "my_param")
        self.assertJE(param, 'schema.type', "string")
        self.assertJE(param, 'required', True)
        self.assertJENotExists(param, 'description')

    def test_path_parameter_description(self):
        param = next(path_parameters(from_doc(
            """Test
            ---
            get:
                parameters extras:
                    my_param: a test description
            """).get('get'), '/path/<my_param>'))
        self.assertJE(param, 'in', "path")
        self.assertJE(param, 'name', "my_param")
        self.assertJE(param, 'schema.type', "string")
        self.assertJE(param, 'required', True)
        self.assertJE(param, 'description', 'a test description')

    def test_path_parameter_extras(self):
        param = next(path_parameters(from_doc(
            """Test
            ---
            get:
                parameters extras:
                    my_param:
                        description: a test description
                        schema:
                            type: number
            """).get('get'), '/path/<my_param>'))
        self.assertJE(param, 'in', "path")
        self.assertJE(param, 'name', "my_param")
        self.assertJE(param, 'schema.type', "number")
        self.assertJE(param, 'required', True)
        self.assertJE(param, 'description', 'a test description')

    def test_make_api_doc(self):
        out = StringIO()
        call_command("makeapidoc", stdout=out)
        self.assertIn(
            "OpenAPI documentation generated: doc/api/openapi.json",
            out.getvalue())
        call_command("makeapidoc", '--format=yaml', stdout=out)
        self.assertIn(
            "OpenAPI documentation generated: doc/api/openapi.yaml",
            out.getvalue())



# Generated by Django 4.2.17 on 2025-04-08 05:19
# flake8: noqa

from django.db import migrations

import pfx.pfxcore.fields.media_field


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='book',
            name='cover',
            field=pfx.pfxcore.fields.media_field.MediaField(blank=True, max_length=255, null=True, verbose_name='Cover'),
        ),
    ]



import logging
from datetime import datetime, timedelta, timezone

from django.contrib.auth import (
    authenticate,
    get_user_model,
    password_validation,
)
from django.contrib.auth.tokens import default_token_generator
from django.core.exceptions import ValidationError
from django.core.mail import EmailMultiAlternatives
from django.core.validators import validate_email
from django.db import transaction
from django.template import loader
from django.utils.decorators import method_decorator
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_decode, urlsafe_base64_encode
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import never_cache
from django.views.decorators.debug import sensitive_post_parameters

import jwt

from pfx.pfxcore.apidoc import Schema, Tag
from pfx.pfxcore.decorator import rest_api, rest_doc, rest_view
from pfx.pfxcore.exceptions import (
    AuthenticationError,
    ForbiddenError,
    NotFoundError,
    UnauthorizedError,
)
from pfx.pfxcore.http import JsonResponse
from pfx.pfxcore.middleware.authentication import JWTTokenDecodeMixin
from pfx.pfxcore.models import CacheableMixin, OtpUserMixin
from pfx.pfxcore.models.login_ban import LoginBan
from pfx.pfxcore.settings import settings
from pfx.pfxcore.shortcuts import delete_token_cookie

from .rest_views import (
    BaseRestView,
    BodyMixin,
    CreateRestViewMixin,
    ModelMixin,
)

logger = logging.getLogger(__name__)
UserModel = get_user_model()
AUTHENTICATION_TAG = Tag("Authentication")


def token_validity(preset):
    presets = dict(
        short=settings.PFX_TOKEN_SHORT_VALIDITY,
        long=settings.PFX_TOKEN_LONG_VALIDITY,
        otp=settings.PFX_TOKEN_OTP_VALIDITY)
    return timedelta(**presets[preset])


@rest_view("/auth")
class AuthenticationView(
        ModelMixin, BodyMixin, JWTTokenDecodeMixin, BaseRestView):
    """The authentication view."""

    model = UserModel
    tags = [AUTHENTICATION_TAG]
    #: The token generator.
    token_generator = default_token_generator

    def login_failed_response(self):
        """Return the response for login failed.

        Can be overridden to customize the error."""
        raise AuthenticationError()

    def login_ban_response(self, ban_dt):
        """Return the response for login temporary ban.

        Can be overridden to customize the error."""
        seconds = int(ban_dt.total_seconds())
        response = JsonResponse(dict(
            message=_(
                "Your connection is temporarily disabled after several "
                "unsuccessful attempts, please retry in {seconds} seconds."
            ).format(seconds=seconds)), status=429)
        response['Retry-after'] = seconds
        return response

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    @rest_api(
        "/login", public=True, method="post", priority_doc=0,
        response_schema='login_schema')
    def login(self, *args, **kwargs):
        """Entrypoint for :code:`POST /login` route.

        Use the request body to authenticate the user using
        the username and password attributes.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        post:
            summary: Login
            description: A login rest services with a `mode` parameter to
                choose between JWT bearer token or cookie authentication.
                In cookie mode, the JWT token is saved in an HTTP-only cookie.
            requestBody:
                content:
                    application/json:
                        schema:
                            properties:
                                username:
                                    type: string
                                password:
                                    type: string
                                    format: password
                                remember_me:
                                    description: If true, use a long validity
                                        token to keep the user logged in
                                        for a long time.
                                    type: boolean
            parameters:
              - in: query
                name: mode
                schema:
                    type: string
                    enum: ['jwt', 'cookie']
                    default: 'jwt'
            responses:
                422:
                    description: The credentials are not valid.
                429:
                    description: Temporary banned after several unsuccessful
                        attempts.
        """
        data = self.deserialize_body()
        username = data.get('username')
        ban_dt = LoginBan.objects.is_ban(username)
        if ban_dt:
            return self.login_ban_response(ban_dt)
        user = authenticate(self.request, username=username,
                            password=data.get('password'))
        if isinstance(user, CacheableMixin):
            user.cache_delete()
        if user is None:
            LoginBan.objects.ban(username)
            return self.login_failed_response()
        LoginBan.objects.unban(username)
        mode = self.request.GET.get('mode', 'jwt')
        remember_me = data.get('remember_me', False)
        if isinstance(user, OtpUserMixin) and user.otp_secret_token:
            return self._login_need_otp_response(user, mode, remember_me)
        return self._login_success(user, mode, remember_me)

    def _login_success(
            self, user, mode, remember_me=False, message=None):
        token = self._prepare_token(user, mode, remember_me)
        if mode == 'cookie':
            if remember_me:
                expires = datetime.now(
                    tz=timezone.utc) + token_validity('long')
            else:
                expires = None  # create a session cookie

            res = JsonResponse(dict(
                need_otp=False,
                user=self.get_user_information(user)))
            res.set_cookie(
                'token', token, secure=settings.PFX_COOKIE_SECURE,
                expires=expires, domain=settings.PFX_COOKIE_DOMAIN,
                httponly=True, samesite=settings.PFX_COOKIE_SAMESITE)
            return res
        return JsonResponse(dict(
            message=message or _("Successful login"),
            need_otp=False,
            token=token,
            user=self.get_user_information(user)))

    def _login_need_otp_response(self, user, mode, remember_me=False):
        """Return the response for login if OTP login is needed.

        Can be overridden to customize the response."""
        token = self._prepare_token(user, mode, remember_me, otp_login=True)
        return JsonResponse(dict(need_otp=True, token=token))

    @method_decorator(never_cache)
    @rest_api(
        "/logout", public=True, method="get", priority_doc=1,
        response_schema='message_schema')
    def logout(self, *args, **kwargs):
        """Entrypoint for :code:`GET /logout` route.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Logout
            description: A service that deletes the authentication cookie
                if it exists.
        """
        return delete_token_cookie(JsonResponse(dict(message="Goodbye")))

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    @rest_api(
        "/change-password", method="post", priority_doc=2,
        response_schema='message_schema')
    def change_password(self, *args, **kwargs):
        """Entrypoint for :code:`POST /change-password` route.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        post:
            summary: Change password
            description: A service to change the password of
                an authenticated user.
            requestBody:
                content:
                    application/json:
                        schema:
                            properties:
                                old_password:
                                    description: the user's current password
                                    type: string
                                    format: password
                                new_password:
                                    description: the user's future password
                                    type: string
                                    format: password
        """
        data = self.deserialize_body()
        user = authenticate(self.request,
                            username=self.request.user.get_username(),
                            password=data.get('old_password'))
        errors = dict()
        try:
            if user is not None and data.get('new_password'):
                password_validation.validate_password(data['new_password'],
                                                      user)
                user.set_password(data.get('new_password'))
                user.save()
                return JsonResponse({
                    'message': _('password updated successfully')
                })
        except ValidationError as e:
            errors['new_password'] = e.error_list
        if not user:
            errors['old_password'] = [_("Incorrect password")]
        if not data.get('new_password'):
            errors.setdefault('new_password', []).append(
                _("Empty password is not allowed"))
        return JsonResponse(
            ValidationError(errors), status=422)

    def _prepare_token(
            self, user, mode='jwt', remember_me=False, otp_login=False,
            **extra_payload):
        exp = datetime.now(tz=timezone.utc) + token_validity(
            otp_login and 'otp' or (remember_me and 'long' or 'short'))
        payload = dict(
            exp=exp,
            pfx_login_options=[mode, remember_me],
            pfx_otp_login=otp_login,
            **self.get_extra_payload(user), **extra_payload)
        return jwt.encode(
            payload,
            user.get_user_jwt_signature_key() + settings.PFX_SECRET_KEY,
            headers=dict(pfx_user_pk=user.pk),
            algorithm="HS256")

    def get_extra_payload(self, user):
        """Get extra payload for user token.

        By default, there is only one private claim in the JWT token
        (exp : expiration). This method can be overridden
        to add claims (key: value attributes) to the JWT token.

        :param user: The user
        :returns: The extra payload
        :rtype: :class:`dict`"""
        return {}

    def get_user_information(self, user):
        """Get the user representation in the view.

        Can be overridden to customize result.

        :param user: The user"""
        info = user.auth_json_repr()
        if isinstance(user, OtpUserMixin):
            info.update(is_otp=user.is_otp)
        return info

    @classmethod
    def get_user_information_schema(cls):
        """Get user representation schema.

        Can be overridden to customize result."""
        return Schema(
            'auth_user', "User", **UserModel.auth_json_repr_schema())

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    @rest_api(
        "/validate-user-token", public=True, method="post", priority_doc=4,
        response_schema='message_schema')
    def validate_user_token(self, *args, **kwargs):
        """Entrypoint for :code:`POST /validate-user-token` route.

        Validate the uid and the set/reset password token contained
        in self.request.body.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        post:
            summary: Validate user token
            description:  Validate the uidb64 and
                the token sent in the set/reset password link.
            requestBody:
                content:
                    application/json:
                        schema:
                            properties:
                                uidb64:
                                    type: string
                                    description: the uid in base64 sent in
                                        the set/reset password link
                                token:
                                    type: string
                                    description: the reset password token
                                        sent in the set/reset password link
        """
        data = self.deserialize_body()
        assert 'uidb64' in data and 'token' in data

        user = self.get_user(data['uidb64'])

        if (user is not None and
                self.token_generator.check_token(user, data['token'])):
            return JsonResponse(
                ValidationError(_('User and token are valid')), status=200)
        return JsonResponse(
            ValidationError(_('User or token is invalid')), status=422)

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    @rest_api(
        "/set-password", public=True, method="post", priority_doc=3,
        response_schema='login_or_message_schema')
    def set_password(self, *args, **kwargs):
        """Entrypoint for :code:`POST /set-password` route.

        Set the password if the base64 uid and the set/reset password
        token are valid. Password, uid and set/reset password token are
        retrieved in self.request.body

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        post:
            summary: Set password
            description: A service for setting the password using a UID and
                a token provided in the email sent by the "forgotten password"
                 or "sign up" services.
            requestBody:
                content:
                    application/json:
                        schema:
                            properties:
                                uidb64:
                                    type: string
                                    description: the uid in base64 sent in
                                        the set/reset password link
                                token:
                                    type: string
                                    description: the reset password token sent
                                        in the set/reset password link
                                password:
                                    description: the new password
                                    type: string
                                    format: password
                                autologin:
                                    description: Automatically logs the user in
                                        upon successful authentication,
                                        utilizing the value of this property
                                        as the login mode.
                                    type: string
                                    enum: ['jwt', 'cookie']
        """
        data = self.deserialize_body()
        assert 'uidb64' in data and 'token' in data

        user = self.get_user(data['uidb64'])

        try:
            if (user is not None and data.get('password') and
                    self.token_generator.check_token(user, data['token'])):
                password_validation.validate_password(data['password'], user)
                user.set_password(data['password'])
                user.is_active = True
                user.save()
                user.refresh_from_db()
                if 'autologin' in data and data['autologin'] in (
                        'jwt', 'cookie'):
                    return self._login_success(user, data['autologin'])
                return JsonResponse({
                    'message': _('password updated successfully')
                })
        except ValidationError as e:
            return JsonResponse(
                ValidationError(dict(password=e.error_list)), status=422)

        if not data.get('password'):
            return JsonResponse(
                ValidationError(dict(
                    password=_("Empty password is not allowed"))), status=422)
        raise UnauthorizedError()

    @method_decorator(never_cache)
    @rest_api("/otp/setup-uri", public=False, method="get", priority_doc=52)
    def otp_setup_uri(self, *args, **kwargs):
        """Entrypoint for :code:`GET /otp/secret-key` route.

        Get the setup uri for the OTP authentication.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        put:
            summary: Get the activation url for the OTP authentication.
            description: This service returns a setup URI to enable the OTP.
                Your front-end application should provide this URI in the form
                of a QR code so that the user can scan it with
                an OTP application. You must then call the /confirm service
                with a valid OTP code to activate OTP authentication.
            responses:
                200:
                    content:
                        application/json:
                            schema:
                                properties:
                                    setup_uri:
                                        type: string
        """
        if not isinstance(self.request.user, OtpUserMixin):
            logger.error("User must inherit OtpUserMixin to activate OTP")
            raise NotFoundError()
        if self.request.user.otp_secret_token:
            return JsonResponse(
                dict(message=_("OTP is already enabled")), status=400)
        self.request.user.enable_otp()
        return JsonResponse(dict(
            setup_uri=self.request.user.get_otp_setup_uri(tmp=True)))

    @method_decorator(never_cache)
    @rest_api(
        "/otp/enable", public=False, method="put", priority_doc=53,
        response_schema='message_schema')
    def otp_enable(self, *args, **kwargs):
        """Entrypoint for :code:`PUT /otp/enable` route.

        Enable the OTP authentication.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        put:
            summary: Enable OTP authentication
            description: A service to enable the OTP authentication with a
                valid OTP code.
            requestBody:
                content:
                    application/json:
                        schema:
                            properties:
                                otp_code:
                                    type: string
                                    description: a valid OTP code
            responses:
                422:
                    description: If the OTP code is not valid.
        """
        if not isinstance(self.request.user, OtpUserMixin):
            logger.error("User must inherit OtpUserMixin to activate OTP")
            raise NotFoundError()
        data = self.deserialize_body()
        if self.request.user.confirm_otp(data.get('otp_code')):
            return self._login_success(
                self.request.user, self.request.login_mode,
                self.request.login_remember_me, message=_("OTP enabled"))
        return JsonResponse(dict(otp_code=[_("Invalid code")]), status=422)

    @method_decorator(never_cache)
    @rest_api("/otp/disable", public=False, method="put", priority_doc=54)
    def otp_disable(self, *args, **kwargs):
        """Entrypoint for :code:`PUT /otp/disable` route.

        Disable the OTP authentication.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        put:
            summary: Disable OTP
            description: A service to disable the OTP authentication.
            requestBody:
                content:
                    application/json:
                        schema:
                            properties:
                                otp_code:
                                    type: string
                                    description: a valid OTP code
            responses:
                422:
                    description: If the OTP code is not valid.
        """
        if not isinstance(self.request.user, OtpUserMixin):
            logger.error("User must inherit OtpUserMixin to activate OTP")
            raise NotFoundError()
        data = self.deserialize_body()
        if self.request.user.is_otp_valid(data.get('otp_code')):
            self.request.user.disable_otp()
            return self._login_success(
                self.request.user, self.request.login_mode,
                self.request.login_remember_me, message=_("OTP disabled"))
        return JsonResponse(dict(otp_code=[_("Invalid code")]), status=422)

    @method_decorator(never_cache)
    @rest_api(
        "/otp/login", public=True, method="post", priority_doc=50,
        response_schema='login_schema')
    def otp_login(self, *args, **kwargs):
        """Entrypoint for :code:`PUT /otp/login` route.

        Login a user with a valid OTP code.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        post:
            summary: OTP login
            description: A login service which validate the OTP code.
            requestBody:
                content:
                    application/json:
                        schema:
                            properties:
                                token:
                                    type: string
                                    description: a valid JWT user token
                                otp_code:
                                    type: string
                                    description: a valid OTP code
            responses:
                422:
                    description: If OTP code is missing, invalid or expired.
                429:
                    description: Temporary banned after several unsuccessful
                        attempts.
                401:
                    description: If the token is missing, invalid or expired.
                403:
                    description: If the OTP is disabled for this user.
        """
        data = self.deserialize_body()
        token = data.get('token', "")
        try:
            user_pk = self.decode_jwt_header(token)
            ban_key = f"otp_{user_pk}"
        except Exception:
            raise UnauthorizedError()
        ban_dt = LoginBan.objects.is_ban(ban_key)
        if ban_dt:
            return self.login_ban_response(ban_dt)
        try:
            user, mode, remember_me = self.decode_jwt(token, otp_login=True)
        except Exception:
            raise UnauthorizedError()
        if not isinstance(user, OtpUserMixin):
            logger.error("User must inherit OtpUserMixin to activate OTP")
            raise NotFoundError()
        if not user.otp_secret_token:
            raise ForbiddenError()
        if user.is_otp_valid(data.get('otp_code')):
            LoginBan.objects.unban(ban_key)
            return self._login_success(user, mode, remember_me)
        LoginBan.objects.ban(ban_key)
        return self.login_failed_response()

    def get_user(self, uidb64):
        """Get user by token

        :param uidb64: The token"""
        try:
            # urlsafe_base64_decode() decodes to bytestring
            uid = urlsafe_base64_decode(uidb64).decode()
            user = UserModel._default_manager.get(pk=uid)
        except (TypeError, ValueError, OverflowError, UserModel.DoesNotExist,
                ValidationError):
            user = None
        return user

    @classmethod
    def generate_schemas(cls):
        """Generate schemas for the class.
        """
        super().generate_schemas()
        cls.login_schema = Schema("login", "Login", properties=dict(
            need_otp=dict(
                type='boolean',
                description="`true` if an OTP login is required "
                "for this user"),
            token=dict(
                type='string',
                description="Returned if `mode` is `'jwt'` or if `need_otp` "
                "is `true` (in this case the token received is only valid "
                "for the OTP login service)"),
            user=dict(
                description="Returned only if `need_otp` is `false`",
                **cls.get_user_information_schema().to_schema())))
        cls.login_or_message_schema = Schema(
            "login_or_message",
            "Login or message (depending autologin parameter)",
            oneOf=[cls.message_schema.id(), cls.login_schema.id()])

    @classmethod
    def get_apidoc_schemas(cls):
        """Get schemas for the class.

        :returns: The schemas list
        :rtype: :class:`list[pfx.pfxcore.apidoc.Schema]`
        """
        return super().get_apidoc_schemas() + [
            cls.login_schema, cls.login_or_message_schema]


class SendMessageTokenMixin:
    """A mixin to send emails with user token."""

    #: The email template.
    email_template_name = None
    #: The email subject template.
    subject_template_name = None
    #: The token generator.
    token_generator = default_token_generator
    #: Extra email context.
    extra_email_context = None
    #: The from value of the email.
    from_email = None
    #: The HTML email template.
    html_email_template_name = None
    #: The email field of user model.
    email_field = 'email'
    #: The language field of user model
    language_field = 'language'

    def send_token_message(self, user):
        """Send an email to a user with a password reset link.

        :param user: The user
        """
        from django.utils import translation
        lang = str(getattr(user, self.language_field, settings.LANGUAGE_CODE))

        token = self.token_generator.make_token(user)
        uidb64 = urlsafe_base64_encode(force_bytes(user.pk))
        data = {
            'target_user': user,
            'token': token,
            'uidb64': uidb64,
            'reset_url': self.reset_url(token, uidb64),
            'site_name': settings.PFX_SITE_NAME,
            'user': user,
            **(self.extra_email_context or {})
        }
        with translation.override(lang):
            subject = loader.render_to_string(self.subject_template_name, data)
            # Email subject *must not* contain newlines
            subject = ''.join(subject.splitlines())
            body = loader.render_to_string(self.email_template_name, data)
            email_message = EmailMultiAlternatives(
                subject, body, self.from_email,
                [getattr(user, self.email_field)])
            if self.html_email_template_name is not None:
                html_email = loader.render_to_string(
                    self.html_email_template_name, data)
                email_message.attach_alternative(
                    html_email, 'text/html')
            email_message.send()

    def reset_url(self, token, uidb64):
        """Get the password reset URL.

        :param token: The token
        :param uidb64: The user PK in base 64
        """
        return settings.PFX_RESET_PASSWORD_URL.format(
            token=token,
            uidb64=uidb64,
        )


@rest_doc("", "post", summary="Signup",
          description="A service that allows visitors to sign up.")
@rest_doc("/meta", "get", summary="Get signup metadata")
@rest_view("/auth/signup")
class SignupView(SendMessageTokenMixin, CreateRestViewMixin, BaseRestView):
    """The view for signup."""
    email_template_name = 'registration/welcome_email.txt'
    subject_template_name = 'registration/welcome_subject.txt'
    token_generator = default_token_generator
    extra_email_context = None
    from_email = None
    html_email_template_name = None
    default_public = True
    model = UserModel
    fields = ['first_name', 'last_name', 'username', 'email']
    tags = [AUTHENTICATION_TAG]

    def validate(self, obj, **kwargs):
        obj.set_unusable_password()
        super().validate(obj, **kwargs)

    def is_valid(self, obj, created=True, **kwargs):
        with transaction.atomic():
            r = super().is_valid(obj, created, **kwargs)
            self.send_token_message(obj)
        return r


@rest_view("/auth")
class ForgottenPasswordView(SendMessageTokenMixin, BodyMixin, BaseRestView):
    """View for forgotten password service."""
    email_template_name = 'registration/password_reset_email.txt'
    subject_template_name = 'registration/password_reset_subject.txt'
    token_generator = default_token_generator
    extra_email_context = None
    from_email = None
    html_email_template_name = None
    #: The ApiDoc tags.
    tags = [AUTHENTICATION_TAG]

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    @rest_api(
        "/forgotten-password", public=True, method="post",
        response_schema='message_schema')
    def forgotten_password(self, *args, **kwargs):
        """Entrypoint for :code:`POST /forgotten-password` route.

        Request a link to reset the password.
        Send an e-mail to the provided email address in the post body
        if it exits with a link to change the user password.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        post:
            summary: Forgotten password
            description: Request an e-mail with a link
                to change the user password.
            requestBody:
                content:
                    application/json:
                        schema:
                            properties:
                                email:
                                    description: the user's email address
                                    type: string
        """
        data = self.deserialize_body()
        email = data.get('email')
        try:
            validate_email(email)
        except ValidationError as e:
            return JsonResponse(
                ValidationError(dict(email=e.error_list)), status=422)
        if email:
            try:
                user = UserModel._default_manager.get(email=email)
            except UserModel.DoesNotExist:
                user = None
            if user is not None:
                self.send_token_message(user)
        return JsonResponse({
            'message': _('If the email address you entered is correct, '
                         'you will receive an email from us with '
                         'instructions to reset your password.')
        })


@rest_view("/auth/otp")
class OtpEmailView(BodyMixin, JWTTokenDecodeMixin, BaseRestView):
    """View for the OTP code email service."""
    #: The email template.
    email_template_name = 'registration/otp_code_email.txt'
    #: The email subject template.
    subject_template_name = 'registration/otp_code_subject.txt'
    #: The HTML email template.
    html_email_template_name = None
    #: Extra email context.
    extra_email_context = None
    #: The from value of the email.
    from_email = None
    #: The email field of user model.
    email_field = 'email'
    #: The language field of user model
    language_field = 'language'
    #: The ApiDoc tags.
    tags = [AUTHENTICATION_TAG]

    @method_decorator(never_cache)
    @rest_api(
        "/email", public=True, method="post", priority_doc=51,
        response_schema='message_schema')
    def sent_email(self, *args, **kwargs):
        """Entrypoint for :code:`POST /email` route.

        Request a new OTP code by email.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        post:
            summary: Send OTP email
            description: Request a new OTP code by email.
            requestBody:
                content:
                    application/json:
                        schema:
                            properties:
                                token:
                                    type: string
                                    description: a valid JWT user token. This
                                        is required only if the user is not
                                        already connected (with a Bearer token
                                        or a Cookie).
            responses:
                401:
                    description: If the token is missing, invalid or expired.
                403:
                    description: If the OTP is disabled for this user.
        """
        if self.request.user.is_anonymous:
            data = self.deserialize_body()
            try:
                user, __, __ = self.decode_jwt(
                    data.get('token', ""), otp_login=True)
            except Exception:
                raise UnauthorizedError()
        else:
            user = self.request.user
        if not isinstance(user, OtpUserMixin):
            logger.error("User must inherit OtpUserMixin to activate OTP")
            raise NotFoundError()
        if not user.otp_secret_token:
            raise ForbiddenError()
        self.send_otp_message(user)
        return JsonResponse({
            'message': _('A new authentication code has been sent by email.')})

    def send_otp_message(self, user):
        """Send an email to a user with an OTP code to a user.

        :param user: The user
        """
        from django.utils import translation
        lang = str(getattr(user, self.language_field, settings.LANGUAGE_CODE))

        otp_code = user.get_hotp_code()
        data = {
            'target_user': user,
            'otp_code': otp_code,
            'otp_validity': settings.PFX_HOTP_CODE_VALIDITY,
            'site_name': settings.PFX_SITE_NAME,
            'user': user,
            **(self.extra_email_context or {})
        }
        with translation.override(lang):
            subject = loader.render_to_string(self.subject_template_name, data)
            # Email subject *must not* contain newlines
            subject = ''.join(subject.splitlines())
            body = loader.render_to_string(self.email_template_name, data)
            email_message = EmailMultiAlternatives(
                subject, body, self.from_email,
                [getattr(user, self.email_field)])
            if self.html_email_template_name is not None:
                html_email = loader.render_to_string(
                    self.html_email_template_name, data)
                email_message.attach_alternative(
                    html_email, 'text/html')
            email_message.send()



from datetime import date
from unittest.mock import patch

import django
from django.test import TestCase
from django.test.utils import override_settings

from pfx.pfxcore.test import (
    APIClient,
    TestAssertMixin,
    format_request,
    get_url,
)
from tests.models import Author


class TestTools(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        cls.author1 = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        pass

    @patch('builtins.print')
    def test_key_error(self, mock_print):
        response = self.client.get('/api/authors/meta/list')
        self.assertRC(response, 200)
        with self.assertRaises(Exception):
            self.assertJE(response, 'items.@0.name-not-exists', 'not exists')

    @patch('builtins.print')
    def test_index_error(self, mock_print):
        response = self.client.get('/api/authors/meta/list')
        self.assertRC(response, 200)
        with self.assertRaises(Exception):
            self.assertJE(response, 'items.@99.name', 'gender')

    @patch('builtins.print')
    def test_print_response(self, mock_print):
        response = self.client.get(f'/api/authors/{self.author1.pk}')

        self.assertRC(response, 200)
        response.print()

        self.assertEqual(mock_print.call_count, 1)
        printed = mock_print.call_args.args[0]
        self.assertIn(
            '******************** http response ********************', printed)
        self.assertIn('Status: 200 OK', printed)
        self.assertIn('Headers: ', printed)
        if django.VERSION[0] >= 4:
            self.assertIn("""  Content-Type: application/json
  Content-Length: %s
  X-Content-Type-Options: nosniff
  Referrer-Policy: same-origin
  Cross-Origin-Opener-Policy: same-origin
  Vary: Accept-Language
  Content-Language: en
""" % (response.headers['Content-Length']), printed)
        else:
            self.assertIn("""  Content-Type: application/json
  Content-Length: %s
  X-Content-Type-Options: nosniff
  Referrer-Policy: same-origin
  Vary: Accept-Language
  Content-Language: en
""" % (response.headers['Content-Length']), printed)
        self.assertIn('Content: ', printed)
        self.assertIn('"books": []', printed)
        self.assertIn('"create_comment": "",', printed)
        self.assertIn(f'"created_at": "{date.today()}"', printed)
        self.assertIn('"first_name": "John Ronald Reuel",', printed)
        self.assertIn(""""gender": {
        "label": "Male",
        "value": "male"
    },""", printed)
        self.assertIn('"last_name": "Tolkien",', printed)
        self.assertIn('"name_length": 25,', printed)
        self.assertIn(f'"pk": {self.author1.pk},', printed)
        self.assertIn('"resource_name": "John Ronald Reuel Tolkien",', printed)
        self.assertIn('"slug": "jrr-tolkien",', printed)
        self.assertIn('"update_comment": ""', printed)
        self.assertIn('"meta": {},', printed)

    @patch('builtins.print')
    @override_settings(PFX_TEST_MODE=True)
    def test_print_request(self, mock_print):
        self.client.post(
            '/api/authors?useless=param&second=param',
            dict(first_name="Arthur Charles",
                 last_name="Clarke"),
            print_request=True)
        self.assertEqual(mock_print.call_count, 1)
        printed = mock_print.call_args.args[0]
        self.assertIn(
            '******************** http request ********************', printed)
        self.assertIn('Path: /api/authors', printed)
        self.assertIn('Method: POST', printed)
        self.assertIn('Query params:', printed)
        self.assertIn('  useless: param', printed)
        self.assertIn('  second: param', printed)
        self.assertIn('Headers: ', printed)
        self.assertIn('  Cookie: ', printed)
        self.assertIn('  Content-Length: 55', printed)
        self.assertIn('  Content-Type: application/json', printed)
        self.assertIn('  X-Custom-Language: en', printed)
        self.assertIn('  X-Print-Request: true', printed)
        self.assertIn('Content: ', printed)
        self.assertIn('"first_name": "Arthur Charles",', printed)
        self.assertIn('"last_name": "Clarke"', printed)

        self.client.get(
            "/api/authors?gender=female", print_request=True)

        self.assertEqual(mock_print.call_count, 2)
        printed = mock_print.call_args.args[0]
        self.assertIn(
            '******************** http request ********************', printed)
        self.assertIn('Path: /api/authors', printed)
        self.assertIn('Method: GET', printed)
        self.assertIn('Query params:', printed)
        self.assertIn('  gender: female', printed)
        self.assertIn('Headers: ', printed)
        self.assertIn('  Cookie: ', printed)
        self.assertIn('  X-Custom-Language: en', printed)
        self.assertIn('  X-Print-Request: true', printed)
        self.assertIn('Content: ', printed)
        self.assertIn('Request is empty', printed)

    @patch('builtins.print')
    @override_settings(PFX_TEST_MODE=True)
    def test_format_request_edge_cases(self, mock_print):
        self.assertEqual(format_request(None), "Request is null")

        self.client.generic(
            method='POST', path='/api/authors',
            data='{"incorrect":"json format',
            content_type='application/json', print_request=True)

        self.assertEqual(mock_print.call_count, 1)
        printed = mock_print.call_args.args[0]
        self.assertIn('Content: ', printed)
        self.assertIn('Json Decode Error', printed)
        self.assertIn("""b'{"incorrect":"json format'""", printed)

        self.client.generic(
            method='POST', path='/api/authors',
            data='Plain text',
            content_type='text/plain', print_request=True)

        self.assertEqual(mock_print.call_count, 2)
        printed = mock_print.call_args.args[0]
        self.assertIn('Content: ', printed)
        self.assertIn("""b'Plain text'""", printed)

    def test_get_url(self):
        self.assertEqual(
            get_url('/test/thing'),
            '/test/thing')
        self.assertEqual(
            get_url('/test/thing', mode="test"),
            '/test/thing?mode=test')
        self.assertEqual(
            get_url('/test/thing', mode="qwertz", uid=12345, other="string"),
            '/test/thing?mode=qwertz&uid=12345&other=string')



from django.http import JsonResponse
from django.test import TestCase, override_settings
from django.urls import include, path

from pfx.pfxcore.decorator.rest import rest_api, rest_doc, rest_view
from pfx.pfxcore.management.commands.makeapidoc import get_spec
from pfx.pfxcore.shortcuts import register_views
from pfx.pfxcore.test import APIClient, TestAssertMixin
from pfx.pfxcore.views import (
    BaseRestView,
    ListRestViewMixin,
    ModelResponseMixin,
    parameters,
)
from tests.models import Author


@rest_view("/list/default")
class DefaultListView(ListRestViewMixin, ModelResponseMixin, BaseRestView):
    model = Author
    default_public = True


@rest_doc("", "get", search=False)
@rest_view("/list/no-search")
class NoSearchListView(ListRestViewMixin, ModelResponseMixin, BaseRestView):
    model = Author
    default_public = True


@rest_doc("", "get", search="A custom description.")
@rest_view("/list/custom-description")
class CustomDescriptionListView(
        ListRestViewMixin, ModelResponseMixin, BaseRestView):
    model = Author
    default_public = True


@rest_view("/custom")
class CustomServiceView(BaseRestView):
    model = Author
    default_public = True

    @rest_api(
        "/default", method="get", parameters=[parameters.groups.List],
        search=True)
    def get_default(self, *args, **kwargs):
        return JsonResponse({})

    @rest_api(
        "/no-search", method="get", parameters=[parameters.groups.List],
        search=False)
    def get_no_search(self, *args, **kwargs):
        return JsonResponse({})

    @rest_api(
        "/custom-description", method="get",
        parameters=[parameters.groups.List],
        search="A custom description.")
    def get_custom_description(self, *args, **kwargs):
        return JsonResponse({})


urlpatterns = [
    path('api/', include(register_views(
        DefaultListView,
        NoSearchListView,
        CustomDescriptionListView,
        CustomServiceView))),
    path('api/', include('pfx.pfxcore.urls'))
]


@override_settings(ROOT_URLCONF=__name__)
class TestAPIDocSearch(TestAssertMixin, TestCase):
    def setUp(self):
        self.client = APIClient(default_locale='en')

    def test_body_to_model(self):
        def get_params(spec, path):
            for p in self.get_val(spec, f'paths.{path}.get.parameters'):
                yield p.get('$ref') or p.get('name'), p

        spec = get_spec(set()).to_dict()

        params = dict(get_params(spec, '/list/default'))
        self.assertJEExists(params, '#/components/parameters/ListSearch')
        self.assertJENotExists(params, 'search')

        params = dict(get_params(spec, '/list/no-search'))
        self.assertJENotExists(params, '#/components/parameters/ListSearch')
        self.assertJENotExists(params, 'search')

        params = dict(get_params(spec, '/list/custom-description'))
        self.assertJENotExists(params, '#/components/parameters/ListSearch')
        self.assertJEExists(params, 'search')
        self.assertJE(params, 'search.description', "A custom description.")

        params = dict(get_params(spec, '/custom/default'))
        self.assertJEExists(params, '#/components/parameters/ListSearch')
        self.assertJENotExists(params, 'search')

        params = dict(get_params(spec, '/custom/no-search'))
        self.assertJENotExists(params, '#/components/parameters/ListSearch')
        self.assertJENotExists(params, 'search')

        params = dict(get_params(spec, '/custom/custom-description'))
        self.assertJENotExists(params, '#/components/parameters/ListSearch')
        self.assertJEExists(params, 'search')
        self.assertJE(params, 'search.description', "A custom description.")



from django.db import connection, models
from django.http import HttpResponse
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import include, path

from pfx.pfxcore import register_views
from pfx.pfxcore.decorator import rest_api, rest_view
from pfx.pfxcore.http import JsonResponse
from pfx.pfxcore.models import JSONReprMixin
from pfx.pfxcore.test import APIClient, TestAssertMixin
from pfx.pfxcore.views import RestView
from tests.models import User
from tests.views import FakeViewMixin


class TestModel(JSONReprMixin, models.Model):
    name = models.CharField("Name", max_length=30)

    class Meta:
        verbose_name = "TestModel"
        verbose_name_plural = "TestModels"
        ordering = ['name', 'pk']

    def __str__(self):
        return self.name


@rest_view("/test")
class TestRestView(FakeViewMixin, RestView):
    default_public = True
    model = TestModel

    @rest_api('/locale', method='get')
    def locale(self, *args, **kwargs):
        return JsonResponse(
            dict(locale=self.request.META.get('HTTP_X_CUSTOM_LANGUAGE')))

    @rest_api('/<int:id>', method='patch')
    def patch(self, *args, **kwargs):
        return self._put(*args, **kwargs)

    @rest_api('', method='options')
    def options(self, *args, **kwargs):
        return HttpResponse(
            headers=dict(
                Allow=', '.join(self._allowed_methods()),
                Content_Length=0
            ))

    @rest_api('', method='head')
    def head(self, *args, **kwargs):
        response = self._get_list(*args, **kwargs)
        response.content = ''
        return response

    @rest_api('', method='trace')
    def trace(self, *args, **kwargs):
        return HttpResponse(content_type='message/http')


urlpatterns = [
    path('api/', include(register_views(TestRestView))),
    path('api/', include('pfx.pfxcore.urls'))
]


@override_settings(ROOT_URLCONF=__name__)
class TestApiClient(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    # @classmethod
    # def setUpClass(cls):

    #     super(TestApiClient, cls).setUpClass()

    @classmethod
    def setUpTestData(cls):
        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(TestModel)
        cls.user = User.objects.create_user(
            username='jrr.tolkien',
            email="jrr.tolkien@oxford.com",
            password='RIGHT PASSWORD',
            first_name='John Ronald Reuel',
            last_name='Tolkien',
        )
        cls.test = TestModel.objects.create(name='test')

    def test_login(self):
        self.assertTrue(
            self.client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD'))
        self.assertTrue(self.client.token)

    def test_logout(self):
        self.client.login(
            username='jrr.tolkien',
            password='RIGHT PASSWORD')
        self.assertTrue(self.client.token)
        self.client.logout()
        self.assertFalse(self.client.token)

    def test_login_cookie(self):
        cookie_client = APIClient(default_locale='en')
        self.assertTrue(
            cookie_client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD'))
        self.assertTrue(cookie_client.token)

    def test_logout_cookie(self):
        cookie_client = APIClient(default_locale='en')
        self.assertTrue(
            cookie_client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD'))
        self.assertTrue(cookie_client.token)
        cookie_client.logout()
        self.assertFalse(cookie_client.token)

    def test_locale(self):
        response = self.client.get(
            '/api/test/locale')
        self.assertRC(response, 200)
        self.assertEqual(response.headers.get('Content-Language'), 'en')

        response = self.client.get(
            '/api/test/locale', locale='fr')
        self.assertRC(response, 200)
        self.assertEqual(response.headers.get('Content-Language'), 'fr')

    def test_get(self):
        response = self.client.get(f'/api/test/{self.test.pk}')
        self.assertRC(response, 200)

    def test_head(self):
        response = self.client.head('/api/test')
        self.assertRC(response, 200)

    def test_post(self):
        response = self.client.post(
            '/api/test', dict(name='test2'))
        self.assertRC(response, 200)

    def test_put(self):
        response = self.client.put(
            f'/api/test/{self.test.pk}', dict(name='test2'))
        self.assertRC(response, 200)

    def test_delete(self):
        response = self.client.delete(
            f'/api/test/{self.test.pk}')
        self.assertRC(response, 200)

    def test_patch(self):
        response = self.client.patch(
            f'/api/test/{self.test.pk}', dict(test='value'))
        self.assertRC(response, 200)

    def test_options(self):
        response = self.client.options(
            '/api/test')
        self.assertRC(response, 200)

    def test_trace(self):
        response = self.client.trace(
            '/api/test')
        self.assertRC(response, 200)



from datetime import date

from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.test import TestCase

from pfx.pfxcore import fields as pfx_fields
from pfx.pfxcore.test import TestAssertMixin
from pfx.pfxcore.views import VF, FieldType, ViewField, ViewModelField
from tests.models import Author, Book
from tests.views import AuthorRestView


class ViewFieldTest(TestAssertMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.author = Author.objects.create(
            first_name='Isaac',
            last_name='Asimov',
            science_fiction=True,
            slug='isaac-asimov')
        cls.author_book1 = Book.objects.create(
            author=cls.author,
            name="The Caves of Steel",
            pub_date=date(1954, 1, 1),
            pages=224,
            rating=4.6)
        cls.author_book2 = Book.objects.create(
            author=cls.author,
            name="The Naked Sun",
            pub_date=date(1957, 1, 1))
        cls.author_book3 = Book.objects.create(
            author=cls.author,
            name="The Robots of Dawn",
            pub_date=date(1983, 1, 1))

    def test_field_type_binding(self):
        self.assertEqual(
            FieldType.from_model_field(models.CharField),
            FieldType.CharField)
        self.assertEqual(
            FieldType.from_model_field(models.SlugField),
            FieldType.CharField)
        self.assertEqual(
            FieldType.from_model_field(models.EmailField),
            FieldType.CharField)
        self.assertEqual(
            FieldType.from_model_field(models.TextField),
            FieldType.TextField)
        self.assertEqual(
            FieldType.from_model_field(models.BooleanField),
            FieldType.BooleanField)
        self.assertEqual(
            FieldType.from_model_field(models.IntegerField),
            FieldType.IntegerField)
        self.assertEqual(
            FieldType.from_model_field(models.FloatField),
            FieldType.FloatField)
        self.assertEqual(
            FieldType.from_model_field(models.DateField),
            FieldType.DateField)
        self.assertEqual(
            FieldType.from_model_field(models.DateTimeField),
            FieldType.DateTimeField)
        self.assertEqual(
            FieldType.from_model_field(models.BooleanField),
            FieldType.BooleanField)
        self.assertEqual(
            FieldType.from_model_field(models.BooleanField),
            FieldType.BooleanField)
        self.assertEqual(
            FieldType.from_model_field(models.BooleanField),
            FieldType.BooleanField)
        self.assertEqual(
            FieldType.from_model_field(models.BooleanField),
            FieldType.BooleanField)
        self.assertEqual(
            FieldType.from_model_field(models.ForeignKey),
            FieldType.ModelObject)
        self.assertEqual(
            FieldType.from_model_field(models.ForeignObjectRel),
            FieldType.ModelObjectList)
        self.assertEqual(
            FieldType.from_model_field(models.JSONField),
            FieldType.JsonObject)
        self.assertEqual(
            FieldType.from_model_field(pfx_fields.MinutesDurationField),
            FieldType.MinutesDurationField)
        self.assertEqual(
            FieldType.from_model_field(pfx_fields.MediaField),
            FieldType.MediaField)

    def test_field_by_name(self):
        field = ViewField.from_name(Book, 'name')
        self.assertTrue(isinstance(field, ViewModelField))
        self.assertEqual(
            field.get_value(self.author_book1), "The Caves of Steel")

        field = ViewField.from_name(Book, 'author__last_name')
        self.assertTrue(isinstance(field, ViewModelField))
        self.assertEqual(field.get_value(self.author_book1), "Asimov")

        # Test with object value
        field = ViewField.from_name(Book, 'author')
        self.assertTrue(isinstance(field, ViewModelField))
        res = field.get_value(self.author_book1)
        self.assertTrue(isinstance(res, Author))
        self.assertEqual(res.pk, self.author.pk)

        # Test with properties
        field = ViewField.from_name(Book, 'author__books_count')
        self.assertFalse(isinstance(field, ViewModelField))
        self.assertTrue(isinstance(field, ViewField))
        self.assertEqual(field.get_value(self.author_book1), 3)

        field = ViewField.from_name(Book, 'author__books_count_prop')
        self.assertFalse(isinstance(field, ViewModelField))
        self.assertTrue(isinstance(field, ViewField))
        self.assertEqual(field.get_value(self.author_book1), 3)

        # Test annotate field
        author = Author.objects.annotate(
            annotate_book_count=models.Count('books')).get(pk=self.author.pk)
        field = ViewField.from_name(Book, 'annotate_book_count')
        self.assertFalse(isinstance(field, ViewModelField))
        self.assertTrue(isinstance(field, ViewField))
        self.assertEqual(field.get_value(author), 3)

        # Check non existent field.
        field = ViewField.from_name(Book, 'doesnotexists')
        with self.assertRaisesRegex(
                AttributeError,
                "'Book' object has no attribute 'doesnotexists'$"):
            field.get_value(self.author_book1)

        # Check non existent object.
        with self.assertRaisesRegex(
                FieldDoesNotExist,
                "Book has no field named 'doesnotexists'$"):
            field = ViewField.from_name(Book, 'doesnotexists__name')

    def test_from_vf(self):
        field = VF('name').to_field(Book)
        self.assertTrue(isinstance(field, ViewModelField))
        self.assertEqual(field.name, 'name')
        self.assertFalse(field.readonly_create)
        self.assertFalse(field.readonly_update)
        self.assertEqual(
            field.get_value(self.author_book1), "The Caves of Steel")

    def test_from_vf_readonly(self):
        field = VF('name', readonly=True).to_field(Book)
        self.assertTrue(isinstance(field, ViewModelField))
        self.assertEqual(field.name, 'name')
        self.assertTrue(field.readonly_create)
        self.assertTrue(field.readonly_update)
        self.assertEqual(
            field.get_value(self.author_book1), "The Caves of Steel")

    def test_from_vf_readonly_create(self):
        field = VF('name', readonly_create=True).to_field(Book)
        self.assertTrue(isinstance(field, ViewModelField))
        self.assertEqual(field.name, 'name')
        self.assertTrue(field.readonly_create)
        self.assertFalse(field.readonly_update)
        self.assertEqual(
            field.get_value(self.author_book1), "The Caves of Steel")

    def test_from_vf_readonly_update(self):
        field = VF('name', readonly_update=True).to_field(Book)
        self.assertTrue(isinstance(field, ViewModelField))
        self.assertEqual(field.name, 'name')
        self.assertFalse(field.readonly_create)
        self.assertTrue(field.readonly_update)
        self.assertEqual(
            field.get_value(self.author_book1), "The Caves of Steel")

    def test_set_value(self):
        view = AuthorRestView()
        view.set_values(self.author_book1, name="The Caves of Steel UPDATED")
        self.assertEqual(
            self.author_book1.name, "The Caves of Steel UPDATED")



from django.db import connection, models
from django.test import TestCase
from django.test.utils import override_settings

from pfx.pfxcore.fields import NH3Field
from pfx.pfxcore.models import JSONReprMixin
from pfx.pfxcore.test import TestAssertMixin


class TestNh3Model(JSONReprMixin, models.Model):
    html = NH3Field()
    custom_html = NH3Field(
        allowed_tags={'h1', 'p', 'span'},
        allowed_attributes={'*': {'class'}})

    class Meta:
        verbose_name = "TestModel"
        verbose_name_plural = "TestModels"
        ordering = ['pk']


@override_settings(ROOT_URLCONF=__name__)
class TestFieldsNh3(TestAssertMixin, TestCase):

    @classmethod
    def setUpTestData(cls):
        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(TestNh3Model)

    def test_save(self):
        t = TestNh3Model.objects.create(
            html="<unknown>AAA</unknown>",
            custom_html='<h1>TEST</h1>'
            '<script>alert("hello")</script>'
            '<p><span class="test" style="font-family: noto;"></span></p>')
        self.assertEqual(t.html, "AAA")
        self.assertEqual(
            t.custom_html, '<h1>TEST</h1><p><span class="test"></span></p>')



import json
from unittest.mock import MagicMock

from django.db import models
from django.test import TestCase

from pfx.pfxcore.exceptions import ModelValidationAPIError
from pfx.pfxcore.models import PFXModelMixin
from pfx.pfxcore.test import TestAssertMixin
from pfx.pfxcore.views import BodyMixin


class CustomModel(PFXModelMixin, models.Model):
    value = models.IntegerField("An integer")

    class Meta:
        managed = False


class CustomView(BodyMixin):
    def __init__(self, **body):
        super().__init__()
        self.request = MagicMock()
        self.request.body = json.dumps(body)


class TestBodyMixin(TestAssertMixin, TestCase):
    def test_body_to_model(self):
        view = CustomView(value=7)
        custom = view.body_to_model(CustomModel)

        self.assertIsInstance(custom, CustomModel)
        self.assertEqual(custom.value, 7)

    def test_body_to_model_ignore_auto_fields(self):
        view = CustomView(id=99999, value=7)
        custom = view.body_to_model(CustomModel)

        self.assertIsInstance(custom, CustomModel)
        self.assertNotEqual(custom.pk, 99999)
        self.assertEqual(custom.value, 7)

    def test_body_to_model_ignore_unknown_fields(self):
        view = CustomView(id=99999, value=7)
        custom = view.body_to_model(CustomModel)

        self.assertIsInstance(custom, CustomModel)
        self.assertNotEqual(custom.pk, 99999)
        self.assertEqual(custom.value, 7)

    def test_body_to_model_validation(self):
        view = CustomView()
        with self.assertRaises(ModelValidationAPIError) as cm:
            view.body_to_model(CustomModel)
        self.assertJE(
            cm.exception.data.message_dict, "value",
            ["This field cannot be null."])

    def test_body_to_model_disable_validation(self):
        view = CustomView()
        custom = view.body_to_model(CustomModel, validate=False)

        self.assertIsInstance(custom, CustomModel)
        self.assertEqual(custom.value, None)



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class ListCount(QueryParameter):
    """Total objects count in response (`meta.count`).

    * if `false` (default value), `meta.count` is not returned
    * if `true`, `meta.count` is returned

    If this attribute is set, `items` will be returned only if
    `items` attribute is explicitly `true`.
    """
    name = 'count'
    schema = dict(type='boolean', default=False)


register_global_parameter(ListCount)



from django.conf import settings
from django.db import models
from django.utils.safestring import mark_safe

import nh3


def get_nh3_default_options():
    nh3_args = {}
    nh3_settings = {
        "NH3_ALLOWED_TAGS": "tags",
        "NH3_ALLOWED_ATTRIBUTES": "attributes",
        "NH3_STRIP_COMMENTS": "strip_comments",
        "NH3_URL_SCHEMES": "url_schemes",
        "NH3_ATTRIBUTE_FILTER": "attribute_filter",
        "NH3_LINK_REL": "link_rel",
        "NH3_GENERIC_ATTRIBUTE_PREFIXES": "generic_attribute_prefixes",
        "NH3_TAG_ATTRIBUTE_VALUES": "tag_attribute_values",
        "NH3_SET_TAG_ATTRIBUTE_VALUES": "set_tag_attribute_values",
    }

    for setting, kwarg in nh3_settings.items():
        if hasattr(settings, setting):
            attr = getattr(settings, setting)
            nh3_args[kwarg] = attr

    return nh3_args


class NH3Field(models.TextField):
    def __init__(
        self,
        allowed_tags=None,
        allowed_attributes=None,
        url_schemes=None,
        strip_comments=None,
        attribute_filter=None,
        link_rel=None,
        generic_attribute_prefixes=None,
        tag_attribute_values=None,
        set_tag_attribute_values=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.nh3_kwargs = get_nh3_default_options()

        if allowed_tags:
            self.nh3_kwargs["tags"] = allowed_tags
        if allowed_attributes:
            self.nh3_kwargs["attributes"] = allowed_attributes
        if url_schemes:
            self.nh3_kwargs["url_schemes"] = url_schemes
        if strip_comments:
            self.nh3_kwargs["strip_comments"] = strip_comments
        if attribute_filter:
            self.nh3_kwargs["attribute_filter"] = attribute_filter
        if link_rel:
            self.nh3_kwargs["link_rel"] = link_rel
        if generic_attribute_prefixes:
            self.nh3_kwargs["generic_attribute_prefixes"] = (
                generic_attribute_prefixes)
        if tag_attribute_values:
            self.nh3_kwargs["tag_attribute_values"] = tag_attribute_values
        if set_tag_attribute_values:
            self.nh3_kwargs["set_tag_attribute_values"] = (
                set_tag_attribute_values)

    def pre_save(self, model_instance, add):
        data = getattr(model_instance, self.attname)
        if data is None:
            return data
        clean_value = nh3.clean(data, **self.nh3_kwargs) if data else ""
        setattr(model_instance, self.attname, mark_safe(clean_value))
        return clean_value

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        # Values are sanitised before saving, so any value returned from the DB
        # is safe to render unescaped.
        return mark_safe(value)



from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings

from pfx.pfxcore.test import TestAssertMixin
from pfx.pfxcore.views import AuthenticationView
from tests.models import User


class TestSettings(TestAssertMixin, TestCase):

    @override_settings(PFX_SECRET_KEY="")
    def test_mandatory_pfx_secret_key(self):
        user = User.objects.create_user(
            username='user',
            email="user@example.com",
            password='test',
            first_name='User',
            last_name='Test')

        with self.assertRaises(ImproperlyConfigured):
            AuthenticationView()._prepare_token(user)



"""
Custom development settings example.

To use custom settting please copy this file:
`cp dev_example.py dev_custom.py to create it.`
"""

from .dev_default import *  # noqa

# SECRET_KEY = 'fake-key'
# INSTALLED_APPS = [
#     'django.contrib.contenttypes',
#     'tests',
# ]
#
# DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'
#
# DATABASES = {
#     'default': {
#         'ENGINE': 'django.db.backends.postgresql_psycopg2',
#         'NAME': 'test-pfx',
#         'USER': 'django',
#         'PASSWORD': 'django',
#         'HOST': 'localhost',
#         'PORT': '5432',
#     }
# }
# ROOT_URLCONF = 'tests.urls'



import logging
import re
from datetime import timedelta

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

logger = logging.getLogger(__name__)


class MinutesDurationField(models.DurationField):
    RE_FLOAT = re.compile(r'^[0-9]*(\.[0-9]*)?$')
    RE_HH_MM = re.compile(r'^([0-9]*):([0-5][0-9])?$')
    RE_HUMAN = re.compile(
        r'^\s*(?:([0-9]*(?:\.[0-9]*)?)h)?\s*(?:([0-9]*)m)?\s*$')
    schema = dict(type='object', properties=dict(
        minutes=dict(type='number', example=90),
        clock_format=dict(type='string', example='01:30'),
        human_format=dict(type='string', example='1h 30m')))

    def to_python(self, value):
        if value is None or value == '':
            return None
        if isinstance(value, timedelta):
            return value
        if isinstance(value, (int, float)):
            return timedelta(hours=value)
        if not isinstance(value, str):
            logger.error(f"invalid value {value} [{type(value)}]")
            raise ValidationError(_("Invalid value."))
        match_float = self.RE_FLOAT.match(value)
        if match_float:
            return timedelta(hours=float(value))
        match_hm = self.RE_HH_MM.match(value)
        if match_hm:
            h, m = match_hm.groups()
            return timedelta(
                hours=h and int(h) or 0, minutes=m and int(m) or 0)
        match_human = self.RE_HUMAN.match(value)
        if match_human:
            h, m = match_human.groups()
            return timedelta(
                hours=h and float(h) or 0, minutes=m and int(m) or 0)
        raise ValidationError(_(
            "Invalid format, it can be a number in hours, “1:05”, “:05”, "
            "“1h 5m”, “1.5h” or “30m”."))

    @staticmethod
    def to_json(value):
        if value is None:
            return None
        minutes = int(value.total_seconds() / 60)
        h, m = minutes // 60, minutes % 60
        return dict(
            minutes=minutes,
            clock_format=f"{minutes // 60}:{minutes % 60:02d}",
            human_format=(
                f'{h and f"{h}h" or ""}\u00A0'
                f'{m and f"{m}m" or ""}'.strip()))



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class MetaFilters(QueryParameter):
    """if `true`, `filters` list is returned in response.

    if one query parameter is set, the default value is `false`.
    """
    name = 'filters'
    schema = dict(type='boolean', default=True)


register_global_parameter(MetaFilters)




class Tag:
    def __init__(self, name, description=None, **kwargs):
        self.name = name
        self.description = description
        self.options = kwargs

    def to_dict(self):
        return dict(
            name=self.name, description=self.description, **self.options)



from decimal import Decimal

from django.db import models


class DecimalField(models.DecimalField):
    def __init__(
            self, *args, percent=False, currency=None,
            json_decimal_places=None, **kw):
        super().__init__(*args, **kw)
        self.percent = percent
        self.currency = currency
        self.json_quantize = Decimal(10) ** -(json_decimal_places or (
            percent and self.decimal_places - 2 or self.decimal_places))

    def to_json_meta(self, meta):
        meta.update(
            percent=self.percent,
            currency=self.currency)
        return meta

    def to_json(self, value):
        if value is None:
            return None
        if self.percent:
            value *= Decimal(100)
        return value.quantize(self.json_quantize)

    def from_json(self, value):
        if value is None:
            return None
        value = Decimal(value)
        return self.percent and value / Decimal(100) or value




from datetime import date
from unittest.mock import MagicMock, patch

from django.db import connection
from django.test import TestCase, override_settings

from pfx.pfxcore.shortcuts import permissions
from pfx.pfxcore.test import APIClient, MockBoto3Client, TestAssertMixin
from tests.models import Author, Book, User


class TestPermissions(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')
        with connection.cursor() as cursor:
            cursor.execute("create extension if not exists unaccent;")

    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user(
            username='user',
            email="user@example.com",
            password='test',
            first_name="Test",
            last_name="User",
            is_superuser=False)
        cls.admin = User.objects.create_user(
            username='admin',
            email="admin@example.com",
            password='test',
            first_name="Test",
            last_name="Admin",
            is_superuser=True)

        cls.author1 = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        cls.author1_book1 = Book.objects.create(
            author=cls.author1,
            name="The Hobbit",
            pub_date=date(1937, 1, 1))
        cls.author2 = Author.objects.create(
            first_name='Philip Kindred',
            last_name='Dick',
            science_fiction=True,
            slug='philip-k-dick')
        cls.author3 = Author.objects.create(
            first_name='Isaac',
            last_name='Asimov',
            science_fiction=True,
            slug='isaac-asimov')

    def test_get_list_admin(self):
        self.client.login(username='admin')

        response = self.client.get('/api/perms/authors?items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

    def test_get_list_user(self):
        self.client.login(username='user')

        response = self.client.get('/api/perms/authors?items=1&count=1')
        self.assertRC(response, 403)

        self.user.user_permissions.add(*permissions('tests.view_author'))

        response = self.client.get('/api/perms/authors?items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

    def test_get_detail_admin(self):
        self.client.login(username='admin')

        response = self.client.get(f'/api/perms/authors/{self.author1.pk}')
        self.assertRC(response, 200)

    def test_get_detail_user(self):
        self.client.login(username='user')

        response = self.client.get(f'/api/perms/authors/{self.author1.pk}')
        self.assertRC(response, 403)

        self.user.user_permissions.add(*permissions('tests.view_author'))

        response = self.client.get(f'/api/perms/authors/{self.author1.pk}')
        self.assertRC(response, 200)

    def test_get_by_slug_admin(self):
        self.client.login(username='admin')

        response = self.client.get(
            f'/api/perms/authors/slug/{self.author1.slug}')
        self.assertRC(response, 200)

    def test_get_by_slug_user(self):
        self.client.login(username='user')

        response = self.client.get(
            f'/api/perms/authors/slug/{self.author1.slug}')
        self.assertRC(response, 403)

        self.user.user_permissions.add(*permissions('tests.view_author'))

        response = self.client.get(
            f'/api/perms/authors/slug/{self.author1.slug}')
        self.assertRC(response, 200)

    def test_create_admin(self):
        self.client.login(username='admin')

        response = self.client.post(
            '/api/perms/authors', dict(
                first_name='New',
                last_name='Author',
                slug='new'))
        self.assertRC(response, 200)

    def test_create_user(self):
        self.client.login(username='user')

        response = self.client.post(
            '/api/perms/authors', dict(
                first_name='New',
                last_name='Author',
                slug='new'))
        self.assertRC(response, 403)

        self.user.user_permissions.add(*permissions('tests.add_author'))

        response = self.client.post(
            '/api/perms/authors', dict(
                first_name='New',
                last_name='Author',
                slug='new'))
        self.assertRC(response, 200)

    def test_update_admin(self):
        self.client.login(username='admin')

        response = self.client.put(
            f'/api/perms/authors/{self.author1.pk}', dict(last_name='UPDATED'))
        self.assertRC(response, 200)

    def test_update_user(self):
        self.client.login(username='user')

        response = self.client.put(
            f'/api/perms/authors/{self.author1.pk}', dict(last_name='UPDATED'))
        self.assertRC(response, 403)

        self.user.user_permissions.add(*permissions('tests.change_author'))

        response = self.client.put(
            f'/api/perms/authors/{self.author1.pk}', dict(last_name='UPDATED'))
        self.assertRC(response, 200)

    def test_delete_admin(self):
        self.client.login(username='admin')

        response = self.client.delete(f'/api/perms/authors/{self.author3.pk}')
        self.assertRC(response, 200)

    def test_delete_user(self):
        self.client.login(username='user')

        response = self.client.delete(f'/api/perms/authors/{self.author3.pk}')
        self.assertRC(response, 403)

        self.user.user_permissions.add(*permissions('tests.delete_author'))

        response = self.client.delete(f'/api/perms/authors/{self.author3.pk}')
        self.assertRC(response, 200)

    @override_settings(
        STORAGE_S3_AWS_REGION="fake-region",
        STORAGE_S3_AWS_ACCESS_KEY="FAKE",
        STORAGE_S3_AWS_SECRET_KEY="FAKE-SECRET",
        STORAGE_S3_AWS_S3_BUCKET="dragonfly.fake",
        STORAGE_S3_AWS_GET_URL_EXPIRE=300,
        STORAGE_S3_AWS_PUT_URL_EXPIRE=300)
    @patch("boto3.client", MagicMock(return_value=MockBoto3Client()))
    def test_media_api_admin(self):
        self.client.login(username='admin')

        response = self.client.get(
            f'/api/perms/books/{self.author1_book1.pk}/cover/upload-url/'
            'cover.png?content-type=image/png')
        self.assertRC(response, 200)

        response = self.client.put(
            f'/api/perms/books/{self.author1_book1.pk}', dict(
                cover=response.json_content['file']))
        self.assertRC(response, 200)

        response = self.client.get(
            f'/api/perms/books/{self.author1_book1.pk}/cover')
        self.assertRC(response, 200)

    @override_settings(
        STORAGE_S3_AWS_REGION="fake-region",
        STORAGE_S3_AWS_ACCESS_KEY="FAKE",
        STORAGE_S3_AWS_SECRET_KEY="FAKE-SECRET",
        STORAGE_S3_AWS_S3_BUCKET="dragonfly.fake",
        STORAGE_S3_AWS_GET_URL_EXPIRE=300,
        STORAGE_S3_AWS_PUT_URL_EXPIRE=300)
    @patch("boto3.client", MagicMock(return_value=MockBoto3Client()))
    def test_media_api_user(self):
        self.client.login(username='user')

        response = self.client.get(
            f'/api/perms/books/{self.author1_book1.pk}/cover/upload-url/'
            'cover.png?content-type=image/png')
        self.assertRC(response, 403)

        self.user.user_permissions.add(*permissions('tests.change_book'))

        response = self.client.get(
            f'/api/perms/books/{self.author1_book1.pk}/cover/upload-url/'
            'cover.png?content-type=image/png')
        self.assertRC(response, 200)

        response = self.client.put(
            f'/api/perms/books/{self.author1_book1.pk}', dict(
                cover=response.json_content['file']))
        self.assertRC(response, 200)

        response = self.client.get(
            f'/api/perms/books/{self.author1_book1.pk}/cover')
        self.assertRC(response, 403)

        self.user.user_permissions.add(*permissions('tests.view_book'))

        response = self.client.get(
            f'/api/perms/books/{self.author1_book1.pk}/cover')
        self.assertRC(response, 200)

    def test_custom_get_admin(self):
        self.client.login(username='admin')

        response = self.client.get('/api/perms/authors/custom')
        self.assertRC(response, 200)

    def test_custom_get_user(self):
        self.client.login(username='user')

        response = self.client.get('/api/perms/authors/custom')
        self.assertRC(response, 403)

        self.user.user_permissions.add(*permissions('tests.view_author'))

        response = self.client.get('/api/perms/authors/custom')
        self.assertRC(response, 200)

    def test_custom_put_admin(self):
        self.client.login(username='admin')

        response = self.client.put('/api/perms/authors/custom')
        self.assertRC(response, 200)

    def test_custom_put_user(self):
        self.client.login(username='user')

        response = self.client.put('/api/perms/authors/custom')
        self.assertRC(response, 403)

        self.user.user_permissions.set(permissions('tests.change_author'))

        response = self.client.put('/api/perms/authors/custom')
        self.assertRC(response, 403)

        self.user.user_permissions.set(permissions(
            'tests.can_customize_author'))

        response = self.client.put('/api/perms/authors/custom')
        self.assertRC(response, 403)

        self.user.user_permissions.set(permissions(
            'tests.change_author', 'tests.can_customize_author'))

        response = self.client.put('/api/perms/authors/custom')
        self.assertRC(response, 200)



from django.db import models
from django.db.models import F, Max, Min


class OrderedModelMixin(models.Model):
    ordered_by = []
    ordered_default = 'last'

    seq = models.PositiveBigIntegerField("Séquence")

    class Meta:
        abstract = True

    @property
    def _ordered_by_values(self):
        return {self.serializable_value(f) for f in self.ordered_by}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__previous_ordered_by_values = self._ordered_by_values

    def clean_fields(self, exclude=None):
        if not self.seq:
            self.seq = self._order_default
        super().clean_fields(exclude=exclude)

    def save(self, *args, **kwargs):
        if not self.seq:
            self.seq = self._order_default
        super().save(*args, **kwargs)
        if self.__previous_ordered_by_values != self._ordered_by_values:
            if self.ordered_default == 'first':
                self.top()
            else:
                self.bottom()

    @property
    def _order_qs(self):
        return self.__class__._default_manager.filter(**{
            f: getattr(self, f) for f in self.ordered_by})

    @property
    def _order_max(self):
        return self._order_qs.aggregate(Max('seq'))['seq__max']

    @property
    def _order_min(self):
        return self._order_qs.aggregate(Min('seq'))['seq__min']

    @property
    def _order_default(self):
        if self.ordered_default == 'first':
            min = self._order_min
            res = min is None and 1 or min - 1
            if res == 0:
                self.first._order_shift()
                return 1
            return res
        else:
            return (self._order_max or 0) + 1

    @property
    def previous(self):
        return self._order_qs.filter(
            seq__lt=self.seq).order_by('seq', 'pk').last()

    @property
    def next(self):
        return self._order_qs.filter(
            seq__gt=self.seq).order_by('seq', 'pk').first()

    @property
    def first(self):
        return self._order_qs.order_by('seq', 'pk').first()

    @property
    def last(self):
        return self._order_qs.order_by('seq', 'pk').last()

    def _order_shift(self):
        self._order_qs.filter(seq__gte=self.seq).update(
            seq=F('seq') + 1)

    def _order_switch(self, other):
        other_seq = other.seq
        other.seq = self.seq
        other.save(update_fields=['seq'])
        self.seq = other_seq
        self.save(update_fields=['seq'])

    def up(self):
        previous = self.previous
        if not previous:
            return
        self._order_switch(previous)

    def down(self):
        next = self.next
        if not next:
            return
        self._order_switch(next)

    def top(self):
        first = self.first
        if first.pk == self.pk:
            return
        seq = first.seq - 1
        if seq == 0:
            first._order_shift()
            seq = 1
        self.seq = seq
        self.save(update_fields=['seq'])

    def bottom(self):
        last = self.last
        if last.pk == self.pk:
            return
        self.seq = last.seq + 1
        self.save(update_fields=['seq'])

    def above(self, other):
        prev = other.previous
        if not prev:
            return self.top()
        seq = other.seq - 1
        if seq == prev.seq:
            other._order_shift()
            seq += 1
        self.seq = seq
        self.save(update_fields=['seq'])

    def below(self, other):
        next = other.next
        if not next:
            return self.bottom()
        seq = other.seq + 1
        if seq == next.seq:
            next._order_shift()
        self.seq = seq
        self.save(update_fields=['seq'])



from . import groups
from .date_format import DateFormat
from .list_count import ListCount
from .list_items import ListItems
from .list_mode import ListMode
from .list_order import ListOrder
from .list_search import ListSearch
from .media_redirect import MediaRedirect
from .meta_fields import MetaFields
from .meta_filters import MetaFilters
from .meta_orders import MetaOrders
from .subset import Subset
from .subset_limit import SubsetLimit
from .subset_offset import SubsetOffset
from .subset_page import SubsetPage
from .subset_page_size import SubsetPageSize
from .subset_page_subset import SubsetPageSubset



from django.urls import include, path

from pfx.pfxcore import register_views

from . import views

handler404 = 'pfx.pfxcore.views.resource_not_found'

urlpatterns = [
    path('api/', include(register_views(
        views.UserRestView))),
    path('api/', include('pfx.pfxcore.urls'))
]



from .common import *  # noqa

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'ci',
        'USER': 'postgres',
        'PASSWORD': 'postgres',
        'HOST': 'postgres',
        'PORT': '5432',
    }
}

TEST_RUNNER = 'xmlrunner.extra.djangotestrunner.XMLTestRunner'
TEST_OUTPUT_FILE_NAME = 'testreport.xml'
TEST_MODE = True
# TEST_OUTPUT_VERBOSE  = 1
# TEST_OUTPUT_DESCRIPTIONS = False



from django.contrib.auth.models import (
    AbstractBaseUser,
    AbstractUser,
    PermissionsMixin,
)
from django.utils.translation import gettext_lazy as _

from .pfx_models import PFXModelMixin


class AbstractPFXBaseUser(PFXModelMixin, AbstractBaseUser):
    """The base abstract user for PFX."""

    class Meta:
        abstract = True

    def auth_json_repr(self, **kw):
        res = self.json_repr(
            username=self.get_username(),
            is_active=self.is_active,
            **kw)
        email_field = self.get_email_field_name()
        if email_field != self.USERNAME_FIELD and hasattr(self, email_field):
            res['email'] = getattr(self, email_field)
        if isinstance(self, PermissionsMixin):
            res['is_superuser'] = self.is_superuser
            res['permissions'] = list(self.get_all_permissions())
        return res

    @classmethod
    def auth_json_repr_schema(cls):
        return cls.json_repr_schema()

    def get_user_jwt_signature_key(self):
        """
        Return a user secret to sign JWT token.

        If not empty, the JWT token validity depends on all values
        user to build the return string. So, each time the returned value
        changes, the previously issued tokens will no longer be valid.
        """
        return self.password


class AbstractPFXUser(AbstractUser, AbstractPFXBaseUser):
    """The base abstract user for PFX with permissions mixin."""

    class Meta:
        verbose_name = _("user")
        verbose_name_plural = _("users")
        abstract = True

    def auth_json_repr(self, **kw):
        res = super().auth_json_repr(
            first_name=self.first_name,
            last_name=self.last_name,
            is_staff=self.is_staff,
            date_joined=self.date_joined,
            **kw)
        return res



# Generated by Django 4.2.17 on 2025-01-21 10:21
# flake8: noqa

from django.db import migrations, models

import pfx.pfxcore.models.pfx_models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('username', models.CharField(max_length=150, unique=True, verbose_name='Username')),
            ],
            options={
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
            },
            bases=(pfx.pfxcore.models.pfx_models.PFXModelMixin, models.Model),
        ),
    ]
